#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.octavia import octavia_base
from heat.engine import support
from heat.engine import translation


class Listener(octavia_base.OctaviaBase):
    """A resource for managing octavia Listeners.

    This resource creates and manages Neutron octavia Listeners,
    which represent a listening endpoint for the vip.
    """

    PROPERTIES = (
        PROTOCOL_PORT, PROTOCOL, LOADBALANCER, DEFAULT_POOL, NAME,
        ADMIN_STATE_UP, DESCRIPTION, DEFAULT_TLS_CONTAINER_REF,
        SNI_CONTAINER_REFS, CONNECTION_LIMIT, TENANT_ID, ALLOWED_CIDRS
    ) = (
        'protocol_port', 'protocol', 'loadbalancer', 'default_pool', 'name',
        'admin_state_up', 'description', 'default_tls_container_ref',
        'sni_container_refs', 'connection_limit', 'tenant_id', 'allowed_cidrs'
    )

    SUPPORTED_PROTOCOLS = (TCP, HTTP, HTTPS, TERMINATED_HTTPS, PROXY, UDP) = (
        'TCP', 'HTTP', 'HTTPS', 'TERMINATED_HTTPS', 'PROXY', 'UDP')

    ATTRIBUTES = (
        LOADBALANCERS_ATTR, DEFAULT_POOL_ID_ATTR
    ) = (
        'loadbalancers', 'default_pool_id'
    )

    properties_schema = {
        PROTOCOL_PORT: properties.Schema(
            properties.Schema.INTEGER,
            _('TCP or UDP port on which to listen for client traffic.'),
            required=True,
            constraints=[
                constraints.Range(1, 65535),
            ]
        ),
        PROTOCOL: properties.Schema(
            properties.Schema.STRING,
            _('Protocol on which to listen for the client traffic.'),
            required=True,
            constraints=[
                constraints.AllowedValues(SUPPORTED_PROTOCOLS),
            ]
        ),
        LOADBALANCER: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the load balancer with which listener '
              'is associated.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('octavia.loadbalancer')
            ]
        ),
        DEFAULT_POOL: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the default pool for the listener.'),
            update_allowed=True,
            constraints=[
                constraints.CustomConstraint('octavia.pool')
            ],
        ),
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of this listener.'),
            update_allowed=True
        ),
        ADMIN_STATE_UP: properties.Schema(
            properties.Schema.BOOLEAN,
            _('The administrative state of this listener.'),
            update_allowed=True,
            default=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of this listener.'),
            update_allowed=True,
            default=''
        ),
        DEFAULT_TLS_CONTAINER_REF: properties.Schema(
            properties.Schema.STRING,
            _('Default TLS container reference to retrieve TLS '
              'information.'),
            update_allowed=True
        ),
        SNI_CONTAINER_REFS: properties.Schema(
            properties.Schema.LIST,
            _('List of TLS container references for SNI.'),
            update_allowed=True
        ),
        CONNECTION_LIMIT: properties.Schema(
            properties.Schema.INTEGER,
            _('The maximum number of connections permitted for this '
              'load balancer. Defaults to -1, which is infinite.'),
            update_allowed=True,
            default=-1,
            constraints=[
                constraints.Range(min=-1),
            ]
        ),
        TENANT_ID: properties.Schema(
            properties.Schema.STRING,
            _('The ID of the tenant who owns the listener.')
        ),
        ALLOWED_CIDRS: properties.Schema(
            properties.Schema.LIST,
            _('A list of IPv4, IPv6 or mix of both CIDRs. The default is all '
              'allowed. When a list of CIDRs is provided, the default '
              'switches to deny all.'),
            update_allowed=True,
            schema=properties.Schema(
                properties.Schema.STRING,
                constraints=[
                    constraints.CustomConstraint('net_cidr')
                ]
            ),
            support_status=support.SupportStatus(version='14.0.0'),
        )
    }

    attributes_schema = {
        LOADBALANCERS_ATTR: attributes.Schema(
            _('ID of the load balancer this listener is associated to.'),
            type=attributes.Schema.LIST
        ),
        DEFAULT_POOL_ID_ATTR: attributes.Schema(
            _('ID of the default pool this listener is associated to.'),
            type=attributes.Schema.STRING
        )
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.LOADBALANCER],
                client_plugin=self.client_plugin(),
                finder='get_loadbalancer',
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.DEFAULT_POOL],
                client_plugin=self.client_plugin(),
                finder='get_pool'
            ),
        ]

    def _prepare_args(self, properties):
        props = dict((k, v) for k, v in properties.items() if v is not None)
        if self.NAME not in props:
            props[self.NAME] = self.physical_resource_name()
        if self.LOADBALANCER in props:
            props['loadbalancer_id'] = props.pop(self.LOADBALANCER)
        if self.DEFAULT_POOL in props:
            props['default_pool_id'] = props.pop(self.DEFAULT_POOL)
        return props

    def validate(self):
        super(Listener, self).validate()
        if self.properties[self.PROTOCOL] == self.TERMINATED_HTTPS:
            if self.properties[self.DEFAULT_TLS_CONTAINER_REF] is None:
                msg = (_('Property %(ref)s required when protocol is '
                       '%(term)s.') % {'ref': self.DEFAULT_TLS_CONTAINER_REF,
                                       'term': self.TERMINATED_HTTPS})
                raise exception.StackValidationFailed(message=msg)

    def _resource_create(self, properties):
        return self.client().listener_create(
            json={'listener': properties})['listener']

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self._update_called = False
        if self.DEFAULT_POOL in prop_diff:
            prop_diff['default_pool_id'] = prop_diff.pop(self.DEFAULT_POOL)
        return prop_diff

    def _resource_update(self, prop_diff):
        self.client().listener_set(self.resource_id,
                                   json={'listener': prop_diff})

    def _resource_delete(self):
        self.client().listener_delete(self.resource_id)

    def _show_resource(self):
        return self.client().listener_show(self.resource_id)


def resource_mapping():
    return {
        'OS::Octavia::Listener': Listener,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from heat.common import exception
from heat.engine import node_data
from heat.engine.resources.openstack.heat import resource_chain
from heat.engine import rsrc_defn
from heat.objects import service as service_objects
from heat.tests import common
from heat.tests import utils

RESOURCE_PROPERTIES = {
    'group': 'test-group',
}

TEMPLATE = {
    'heat_template_version': '2016-04-08',
    'resources': {
        'test-chain': {
            'type': 'OS::Heat::ResourceChain',
            'properties': {
                'resources': ['OS::Heat::SoftwareConfig',
                              'OS::Heat::StructuredConfig'],
                'concurrent': False,
                'resource_properties': RESOURCE_PROPERTIES,
            }
        }
    }
}


class ResourceChainTest(common.HeatTestCase):

    def setUp(self):
        super(ResourceChainTest, self).setUp()

        self.stack = None  # hold on to stack to prevent weakref cleanup

    def test_child_template_without_concurrency(self):
        # Test
        chain = self._create_chain(TEMPLATE)
        child_template = chain.child_template()

        # Verify
        tmpl = child_template.t
        self.assertEqual('2015-04-30', tmpl['heat_template_version'])
        self.assertEqual(2, len(child_template.t['resources']))

        resource = tmpl['resources']['0']
        self.assertEqual('OS::Heat::SoftwareConfig', resource['type'])
        self.assertEqual(RESOURCE_PROPERTIES, resource['properties'])
        self.assertNotIn('depends_on', resource)

        resource = tmpl['resources']['1']
        self.assertEqual('OS::Heat::StructuredConfig', resource['type'])
        self.assertEqual(RESOURCE_PROPERTIES, resource['properties'])
        self.assertEqual(['0'], resource['depends_on'])

    @mock.patch.object(service_objects.Service, 'active_service_count')
    def test_child_template_with_concurrent(self, mock_count):

        # Setup
        tmpl_def = copy.deepcopy(TEMPLATE)
        tmpl_def['resources']['test-chain']['properties']['concurrent'] = True
        chain = self._create_chain(tmpl_def)
        mock_count.return_value = 5

        # Test
        child_template = chain.child_template()

        # Verify
        # Trimmed down version of above that just checks the depends_on
        # isn't present
        tmpl = child_template.t
        resource = tmpl['resources']['0']
        self.assertNotIn('depends_on', resource)

        resource = tmpl['resources']['1']
        self.assertNotIn('depends_on', resource)

    @mock.patch.object(service_objects.Service, 'active_service_count')
    def test_child_template_with_concurrent_limit(self, mock_count):

        tmpl_def = copy.deepcopy(TEMPLATE)
        tmpl_def['resources']['test-chain']['properties']['concurrent'] = True
        tmpl_def['resources']['test-chain']['properties']['resources'] = [
            'OS::Heat::SoftwareConfig', 'OS::Heat::StructuredConfig',
            'OS::Heat::SoftwareConfig', 'OS::Heat::StructuredConfig']
        chain = self._create_chain(tmpl_def)
        mock_count.return_value = 2

        child_template = chain.child_template()

        tmpl = child_template.t
        resource = tmpl['resources']['0']
        self.assertNotIn('depends_on', resource)

        resource = tmpl['resources']['1']
        self.assertNotIn('depends_on', resource)

        resource = tmpl['resources']['2']
        self.assertEqual(['0'], resource['depends_on'])

        resource = tmpl['resources']['3']
        self.assertEqual(['1'], resource['depends_on'])

    def test_child_template_default_concurrent(self):
        # Setup
        tmpl_def = copy.deepcopy(TEMPLATE)
        tmpl_def['resources']['test-chain']['properties'].pop('concurrent')
        chain = self._create_chain(tmpl_def)

        # Test
        child_template = chain.child_template()

        # Verify
        # Trimmed down version of above that just checks the depends_on
        # isn't present
        tmpl = child_template.t
        resource = tmpl['resources']['0']
        self.assertNotIn('depends_on', resource)

        resource = tmpl['resources']['1']
        self.assertEqual(['0'], resource['depends_on'])

    def test_child_template_empty_resource_list(self):
        # Setup
        tmpl_def = copy.deepcopy(TEMPLATE)
        tmpl_def['resources']['test-chain']['properties']['resources'] = []
        chain = self._create_chain(tmpl_def)

        # Test
        child_template = chain.child_template()

        # Verify
        tmpl = child_template.t
        # No error, but no resources to create
        self.assertNotIn('resources', tmpl)
        # Sanity check that it's actually a template
        self.assertIn('heat_template_version', tmpl)

    def test_validate_nested_stack(self):
        # Test - should not raise exception
        chain = self._create_chain(TEMPLATE)
        chain.validate_nested_stack()

    def test_validate_reference_attr_with_none_ref(self):
        chain = self._create_chain(TEMPLATE)
        self.patchobject(chain, 'referenced_attrs',
                         return_value=set([('config', None)]))
        self.assertIsNone(chain.validate())

    def test_validate_incompatible_properties(self):
        # Tests a resource in the chain that does not support the properties
        # specified to each resource.

        # Setup
        tmpl_def = copy.deepcopy(TEMPLATE)
        tmpl_res_prop = tmpl_def['resources']['test-chain']['properties']
        res_list = tmpl_res_prop['resources']
        res_list.append('OS::Heat::RandomString')

        # Test
        chain = self._create_chain(tmpl_def)

        try:
            chain.validate_nested_stack()
            self.fail('Exception expected')
        except exception.StackValidationFailed as e:
            self.assertEqual('property error: '
                             'resources.test<nested_stack>.resources[2].'
                             'properties: unknown property group',
                             e.message.lower())

    def test_validate_fake_resource_type(self):
        # Setup
        tmpl_def = copy.deepcopy(TEMPLATE)
        tmpl_res_prop = tmpl_def['resources']['test-chain']['properties']
        res_list = tmpl_res_prop['resources']
        res_list.append('foo')

        # Test
        chain = self._create_chain(tmpl_def)

        try:
            chain.validate_nested_stack()
            self.fail('Exception expected')
        except exception.StackValidationFailed as e:
            self.assertIn('could not be found', e.message.lower())
            self.assertIn('foo', e.message)

    @mock.patch.object(resource_chain.ResourceChain, 'create_with_template')
    def test_handle_create(self, mock_create):
        # Tests the handle create is propagated upwards with the
        # child template.

        # Setup
        chain = self._create_chain(TEMPLATE)

        # Test
        chain.handle_create()

        # Verify
        expected_tmpl = chain.child_template()
        mock_create.assert_called_once_with(expected_tmpl)

    @mock.patch.object(resource_chain.ResourceChain, 'update_with_template')
    def test_handle_update(self, mock_update):
        # Test the handle update is propagated upwards with the child
        # template.

        # Setup
        chain = self._create_chain(TEMPLATE)

        # Test
        json_snippet = rsrc_defn.ResourceDefinition(
            'test-chain', 'OS::Heat::ResourceChain',
            TEMPLATE['resources']['test-chain']['properties'])

        chain.handle_update(json_snippet, None, None)

        # Verify
        expected_tmpl = chain.child_template()
        mock_update.assert_called_once_with(expected_tmpl)

    def test_child_params(self):
        chain = self._create_chain(TEMPLATE)
        self.assertEqual({}, chain.child_params())

    def _create_chain(self, t):
        self.stack = utils.parse_stack(t)
        snip = self.stack.t.resource_definitions(self.stack)['test-chain']
        chain = resource_chain.ResourceChain('test', snip, self.stack)
        return chain

    def test_get_attribute_convg(self):
        cache_data = {'test-chain': node_data.NodeData.from_dict({
            'uuid': mock.ANY,
            'id': mock.ANY,
            'action': 'CREATE',
            'status': 'COMPLETE',
            'attrs': {'refs': ['rsrc1', 'rsrc2']}
        })}
        stack = utils.parse_stack(TEMPLATE, cache_data=cache_data)
        rsrc = stack.defn['test-chain']
        self.assertEqual(['rsrc1', 'rsrc2'], rsrc.FnGetAtt('refs'))


class ResourceChainAttrTest(common.HeatTestCase):
    def test_aggregate_attribs(self):
        """Test attribute aggregation.

        Test attribute aggregation and that we mimic the nested resource's
        attributes.
        """
        chain = self._create_dummy_stack()
        expected = ['0', '1']
        self.assertEqual(expected, chain.FnGetAtt('foo'))
        self.assertEqual(expected, chain.FnGetAtt('Foo'))

    def test_index_dotted_attribs(self):
        """Test attribute aggregation.

        Test attribute aggregation and that we mimic the nested resource's
        attributes.
        """
        chain = self._create_dummy_stack()
        self.assertEqual('0', chain.FnGetAtt('resource.0.Foo'))
        self.assertEqual('1', chain.FnGetAtt('resource.1.Foo'))

    def test_index_path_attribs(self):
        """Test attribute aggregation.

        Test attribute aggregation and that we mimic the nested resource's
        attributes.
        """
        chain = self._create_dummy_stack()
        self.assertEqual('0', chain.FnGetAtt('resource.0', 'Foo'))
        self.assertEqual('1', chain.FnGetAtt('resource.1', 'Foo'))

    def test_index_deep_path_attribs(self):
        """Test attribute aggregation.

        Test attribute aggregation and that we mimic the nested resource's
        attributes.
        """
        chain = self._create_dummy_stack(expect_attrs={'0': 2, '1': 3})
        self.assertEqual(2, chain.FnGetAtt('resource.0',
                                           'nested_dict', 'dict', 'b'))
        self.assertEqual(3, chain.FnGetAtt('resource.1',
                                           'nested_dict', 'dict', 'b'))

    def test_aggregate_deep_path_attribs(self):
        """Test attribute aggregation.

        Test attribute aggregation and that we mimic the nested resource's
        attributes.
        """
        chain = self._create_dummy_stack(expect_attrs={'0': 3, '1': 3})
        expected = [3, 3]
        self.assertEqual(expected, chain.FnGetAtt('nested_dict', 'list', 2))

    def test_aggregate_refs(self):
        """Test resource id aggregation."""
        chain = self._create_dummy_stack()
        expected = ['ID-0', 'ID-1']
        self.assertEqual(expected, chain.FnGetAtt("refs"))

    def test_aggregate_refs_with_index(self):
        """Test resource id aggregation with index."""
        chain = self._create_dummy_stack()
        expected = ['ID-0', 'ID-1']
        self.assertEqual(expected[0], chain.FnGetAtt("refs", 0))
        self.assertEqual(expected[1], chain.FnGetAtt("refs", 1))
        self.assertIsNone(chain.FnGetAtt("refs", 2))

    def test_aggregate_outputs(self):
        """Test outputs aggregation."""
        expected = {'0': ['foo', 'bar'], '1': ['foo', 'bar']}
        chain = self._create_dummy_stack(expect_attrs=expected)
        self.assertEqual(expected, chain.FnGetAtt('attributes', 'list'))

    def test_aggregate_outputs_no_path(self):
        """Test outputs aggregation with missing path."""
        chain = self._create_dummy_stack()
        self.assertRaises(exception.InvalidTemplateAttribute,
                          chain.FnGetAtt, 'attributes')

    def test_index_refs(self):
        """Tests getting ids of individual resources."""
        chain = self._create_dummy_stack()
        self.assertEqual("ID-0", chain.FnGetAtt('resource.0'))
        self.assertEqual("ID-1", chain.FnGetAtt('resource.1'))
        ex = self.assertRaises(exception.NotFound, chain.FnGetAtt,
                               'resource.2')
        self.assertIn("Member '2' not found in group resource 'test'",
                      str(ex))

    def _create_dummy_stack(self, expect_count=2, expect_attrs=None):
        self.stack = utils.parse_stack(TEMPLATE)
        snip = self.stack.t.resource_definitions(self.stack)['test-chain']
        chain = resource_chain.ResourceChain('test', snip, self.stack)
        attrs = {}
        refids = {}
        if expect_attrs is None:
            expect_attrs = {}
        for index in range(expect_count):
            res = str(index)
            attrs[index] = expect_attrs.get(res, res)
            refids[index] = 'ID-%s' % res

        names = [str(name) for name in range(expect_count)]
        chain._resource_names = mock.Mock(return_value=names)
        self._stub_get_attr(chain, refids, attrs)
        return chain

    def _stub_get_attr(self, chain, refids, attrs):
        def ref_id_fn(res_name):
            return refids[int(res_name)]

        def attr_fn(args):
            res_name = args[0]
            return attrs[int(res_name)]

        def get_output(output_name):
            outputs = chain._nested_output_defns(chain._resource_names(),
                                                 attr_fn, ref_id_fn)
            op_defns = {od.name: od for od in outputs}
            if output_name not in op_defns:
                raise exception.NotFound('Specified output key %s not found.' %
                                         output_name)
            return op_defns[output_name].get_value()

        orig_get_attr = chain.FnGetAtt

        def get_attr(attr_name, *path):
            if not path:
                attr = attr_name
            else:
                attr = (attr_name,) + path
            # Mock referenced_attrs() so that _nested_output_definitions()
            # will include the output required for this attribute
            chain.referenced_attrs = mock.Mock(return_value=[attr])

            # Pass through to actual function under test
            return orig_get_attr(attr_name, *path)

        chain.FnGetAtt = mock.Mock(side_effect=get_attr)
        chain.get_output = mock.Mock(side_effect=get_output)


class ResourceChainAttrFallbackTest(ResourceChainAttrTest):
    def _stub_get_attr(self, chain, refids, attrs):
        # Raise NotFound when getting output, to force fallback to old-school
        # grouputils functions
        chain.get_output = mock.Mock(side_effect=exception.NotFound)

        def make_fake_res(idx):
            fr = mock.Mock()
            fr.stack = chain.stack
            fr.FnGetRefId.return_value = refids[idx]
            fr.FnGetAtt.return_value = attrs[idx]
            return fr

        fake_res = {str(i): make_fake_res(i) for i in refids}
        chain.nested = mock.Mock(return_value=fake_res)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import urllib.error
import urllib.request

from oslo_config import cfg
import requests
from requests import exceptions

from heat.common import urlfetch
from heat.tests import common


class Response(object):
    def __init__(self, buf=''):
        self.buf = buf

    def iter_content(self, chunk_size=1):
        while self.buf:
            yield self.buf[:chunk_size]
            self.buf = self.buf[chunk_size:]

    def raise_for_status(self):
        pass


class UrlFetchTest(common.HeatTestCase):

    def test_file_scheme_default_behaviour(self):
        self.assertRaises(urlfetch.URLFetchError,
                          urlfetch.get, 'file:///etc/profile')

    def test_file_scheme_supported(self):
        data = '{ "foo": "bar" }'
        url = 'file:///etc/profile'
        mock_open = self.patchobject(urllib.request, 'urlopen')
        mock_open.return_value = io.StringIO(data)
        self.assertEqual(data, urlfetch.get(url, allowed_schemes=['file']))
        mock_open.assert_called_once_with(url)

    def test_file_scheme_failure(self):
        url = 'file:///etc/profile'
        mock_open = self.patchobject(urllib.request, 'urlopen')
        mock_open.side_effect = urllib.error.URLError('oops')
        self.assertRaises(urlfetch.URLFetchError,
                          urlfetch.get, url, allowed_schemes=['file'])
        mock_open.assert_called_once_with(url)

    def test_http_scheme(self):
        url = 'http://example.com/template'
        data = b'{ "foo": "bar" }'
        response = Response(data)
        mock_get = self.patchobject(requests, 'get')
        mock_get.return_value = response
        self.assertEqual(data, urlfetch.get(url))
        mock_get.assert_called_once_with(url, stream=True, timeout=60)

    def test_https_scheme(self):
        url = 'https://example.com/template'
        data = b'{ "foo": "bar" }'
        response = Response(data)
        mock_get = self.patchobject(requests, 'get')
        mock_get.return_value = response
        self.assertEqual(data, urlfetch.get(url))
        mock_get.assert_called_once_with(url, stream=True, timeout=60)

    def test_http_error(self):
        url = 'http://example.com/template'
        mock_get = self.patchobject(requests, 'get')
        mock_get.side_effect = exceptions.HTTPError()
        self.assertRaises(urlfetch.URLFetchError, urlfetch.get, url)
        mock_get.assert_called_once_with(url, stream=True, timeout=60)

    def test_non_exist_url(self):
        url = 'http://non-exist.com/template'
        mock_get = self.patchobject(requests, 'get')
        mock_get.side_effect = exceptions.Timeout()
        self.assertRaises(urlfetch.URLFetchError, urlfetch.get, url)
        mock_get.assert_called_once_with(url, stream=True, timeout=60)

    def test_garbage(self):
        self.assertRaises(urlfetch.URLFetchError, urlfetch.get, 'wibble')

    def test_max_fetch_size_okay(self):
        url = 'http://example.com/template'
        data = b'{ "foo": "bar" }'
        response = Response(data)
        cfg.CONF.set_override('max_template_size', 500)
        mock_get = self.patchobject(requests, 'get')
        mock_get.return_value = response
        urlfetch.get(url)
        mock_get.assert_called_once_with(url, stream=True, timeout=60)

    def test_max_fetch_size_error(self):
        url = 'http://example.com/template'
        data = b'{ "foo": "bar" }'
        response = Response(data)
        cfg.CONF.set_override('max_template_size', 5)
        mock_get = self.patchobject(requests, 'get')
        mock_get.return_value = response
        exception = self.assertRaises(urlfetch.URLFetchError,
                                      urlfetch.get, url)
        self.assertIn("Template exceeds", str(exception))
        mock_get.assert_called_once_with(url, stream=True, timeout=60)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial revision

Revision ID: c6214ca60943
Revises:
Create Date: 2023-03-22 18:04:02.387269
"""

from alembic import op
import sqlalchemy as sa

import heat.db.types

# revision identifiers, used by Alembic.
revision = 'c6214ca60943'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        'raw_template_files',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('files', heat.db.types.Json(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )
    op.create_table(
        'resource_properties_data',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('data', heat.db.types.Json(), nullable=True),
        sa.Column('encrypted', sa.Boolean(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )
    op.create_table(
        'service',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('engine_id', sa.String(length=36), nullable=False),
        sa.Column('host', sa.String(length=255), nullable=False),
        sa.Column('hostname', sa.String(length=255), nullable=False),
        sa.Column('binary', sa.String(length=255), nullable=False),
        sa.Column('topic', sa.String(length=255), nullable=False),
        sa.Column('report_interval', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )
    op.create_table(
        'software_config',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('group', sa.String(length=255), nullable=True),
        sa.Column('config', heat.db.types.Json(), nullable=True),
        sa.Column('tenant', sa.String(length=64), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )
    op.create_index(
        op.f('ix_software_config_tenant'),
        'software_config',
        ['tenant'],
        unique=False,
    )
    op.create_table(
        'user_creds',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('username', sa.String(length=255), nullable=True),
        sa.Column('password', sa.String(length=255), nullable=True),
        sa.Column('region_name', sa.String(length=255), nullable=True),
        sa.Column('decrypt_method', sa.String(length=64), nullable=True),
        sa.Column('tenant', sa.String(length=1024), nullable=True),
        sa.Column('auth_url', sa.Text(), nullable=True),
        sa.Column('tenant_id', sa.String(length=256), nullable=True),
        sa.Column('trust_id', sa.String(length=255), nullable=True),
        sa.Column('trustor_user_id', sa.String(length=64), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )
    op.create_table(
        'raw_template',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('template', heat.db.types.Json(), nullable=True),
        sa.Column('files', heat.db.types.Json(), nullable=True),
        sa.Column(
            'environment', heat.db.types.Json(), nullable=True
        ),
        sa.Column('files_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['files_id'],
            ['raw_template_files.id'],
            name='raw_tmpl_files_fkey_ref',
        ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )
    op.create_table(
        'software_deployment',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('server_id', sa.String(length=36), nullable=False),
        sa.Column('config_id', sa.String(length=36), nullable=False),
        sa.Column(
            'input_values', heat.db.types.Json(), nullable=True
        ),
        sa.Column(
            'output_values', heat.db.types.Json(), nullable=True
        ),
        sa.Column('action', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=255), nullable=True),
        sa.Column('status_reason', sa.Text(), nullable=True),
        sa.Column('tenant', sa.String(length=64), nullable=False),
        sa.Column(
            'stack_user_project_id', sa.String(length=64), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ['config_id'],
            ['software_config.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
    )
    op.create_index(
        'ix_software_deployment_created_at',
        'software_deployment',
        ['created_at'],
        unique=False,
    )
    op.create_index(
        op.f('ix_software_deployment_server_id'),
        'software_deployment',
        ['server_id'],
        unique=False,
    )
    op.create_index(
        op.f('ix_software_deployment_tenant'),
        'software_deployment',
        ['tenant'],
        unique=False,
    )
    op.create_table(
        'stack',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('raw_template_id', sa.Integer(), nullable=False),
        sa.Column('prev_raw_template_id', sa.Integer(), nullable=True),
        sa.Column('user_creds_id', sa.Integer(), nullable=True),
        sa.Column('username', sa.String(length=256), nullable=True),
        sa.Column('owner_id', sa.String(length=36), nullable=True),
        sa.Column('action', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=255), nullable=True),
        sa.Column('status_reason', sa.Text(), nullable=True),
        sa.Column('timeout', sa.Integer(), nullable=True),
        sa.Column('tenant', sa.String(length=256), nullable=True),
        sa.Column('disable_rollback', sa.Boolean(), nullable=False),
        sa.Column(
            'stack_user_project_id', sa.String(length=64), nullable=True
        ),
        sa.Column('backup', sa.Boolean(), nullable=True),
        sa.Column('nested_depth', sa.Integer(), nullable=True),
        sa.Column('convergence', sa.Boolean(), nullable=True),
        sa.Column('current_traversal', sa.String(length=36), nullable=True),
        sa.Column(
            'current_deps', heat.db.types.Json(), nullable=True
        ),
        sa.Column(
            'parent_resource_name', sa.String(length=255), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ['prev_raw_template_id'],
            ['raw_template.id'],
        ),
        sa.ForeignKeyConstraint(
            ['raw_template_id'],
            ['raw_template.id'],
        ),
        sa.ForeignKeyConstraint(
            ['user_creds_id'],
            ['user_creds.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
    )
    op.create_index(
        'ix_stack_name', 'stack', ['name'], unique=False, mysql_length=255
    )
    op.create_index(
        'ix_stack_tenant', 'stack', ['tenant'], unique=False, mysql_length=255
    )
    op.create_index(
        op.f('ix_stack_owner_id'), 'stack', ['owner_id'], unique=False
    )
    op.create_table(
        'event',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('stack_id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('resource_action', sa.String(length=255), nullable=True),
        sa.Column('resource_status', sa.String(length=255), nullable=True),
        sa.Column('resource_name', sa.String(length=255), nullable=True),
        sa.Column(
            'physical_resource_id', sa.String(length=255), nullable=True
        ),
        sa.Column(
            'resource_status_reason', sa.String(length=255), nullable=True
        ),
        sa.Column('resource_type', sa.String(length=255), nullable=True),
        sa.Column('resource_properties', sa.PickleType(), nullable=True),
        sa.Column('rsrc_prop_data_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['rsrc_prop_data_id'],
            ['resource_properties_data.id'],
            name='ev_rsrc_prop_data_ref',
        ),
        sa.ForeignKeyConstraint(
            ['stack_id'],
            ['stack.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid'),
        mysql_engine='InnoDB',
    )
    op.create_table(
        'resource',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('nova_instance', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('action', sa.String(length=255), nullable=True),
        sa.Column('status', sa.String(length=255), nullable=True),
        sa.Column('status_reason', sa.Text(), nullable=True),
        sa.Column('stack_id', sa.String(length=36), nullable=False),
        sa.Column(
            'rsrc_metadata', heat.db.types.Json(), nullable=True
        ),
        sa.Column(
            'properties_data', heat.db.types.Json(), nullable=True
        ),
        sa.Column('engine_id', sa.String(length=36), nullable=True),
        sa.Column('atomic_key', sa.Integer(), nullable=True),
        sa.Column('needed_by', heat.db.types.List(), nullable=True),
        sa.Column('requires', heat.db.types.List(), nullable=True),
        sa.Column('replaces', sa.Integer(), nullable=True),
        sa.Column('replaced_by', sa.Integer(), nullable=True),
        sa.Column('current_template_id', sa.Integer(), nullable=True),
        sa.Column('properties_data_encrypted', sa.Boolean(), nullable=True),
        sa.Column('root_stack_id', sa.String(length=36), nullable=True),
        sa.Column('rsrc_prop_data_id', sa.Integer(), nullable=True),
        sa.Column('attr_data_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['attr_data_id'],
            ['resource_properties_data.id'],
            name='rsrc_attr_data_ref',
        ),
        sa.ForeignKeyConstraint(
            ['current_template_id'],
            ['raw_template.id'],
        ),
        sa.ForeignKeyConstraint(
            ['rsrc_prop_data_id'],
            ['resource_properties_data.id'],
            name='rsrc_rsrc_prop_data_ref',
        ),
        sa.ForeignKeyConstraint(
            ['stack_id'],
            ['stack.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid'),
        mysql_engine='InnoDB',
    )
    op.create_index(
        op.f('ix_resource_root_stack_id'),
        'resource',
        ['root_stack_id'],
        unique=False,
    )
    op.create_table(
        'snapshot',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('stack_id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('status', sa.String(length=255), nullable=True),
        sa.Column('status_reason', sa.String(length=255), nullable=True),
        sa.Column('data', heat.db.types.Json(), nullable=True),
        sa.Column('tenant', sa.String(length=64), nullable=False),
        sa.ForeignKeyConstraint(
            ['stack_id'],
            ['stack.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )
    op.create_index(
        op.f('ix_snapshot_tenant'), 'snapshot', ['tenant'], unique=False
    )
    op.create_table(
        'stack_lock',
        sa.Column('stack_id', sa.String(length=36), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('engine_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(
            ['stack_id'],
            ['stack.id'],
        ),
        sa.PrimaryKeyConstraint('stack_id'),
        mysql_engine='InnoDB',
    )
    op.create_table(
        'stack_tag',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('tag', sa.Unicode(length=80), nullable=True),
        sa.Column('stack_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(
            ['stack_id'],
            ['stack.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )
    op.create_table(
        'sync_point',
        sa.Column('entity_id', sa.String(length=36), nullable=False),
        sa.Column('traversal_id', sa.String(length=36), nullable=False),
        sa.Column('is_update', sa.Boolean(), nullable=False),
        sa.Column('atomic_key', sa.Integer(), nullable=False),
        sa.Column('stack_id', sa.String(length=36), nullable=False),
        sa.Column(
            'input_data', heat.db.types.Json(), nullable=True
        ),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ['stack_id'],
            ['stack.id'],
        ),
        sa.PrimaryKeyConstraint('entity_id', 'traversal_id', 'is_update'),
    )
    op.create_table(
        'resource_data',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('key', sa.String(length=255), nullable=True),
        sa.Column('value', sa.Text(), nullable=True),
        sa.Column('redact', sa.Boolean(), nullable=True),
        sa.Column('decrypt_method', sa.String(length=64), nullable=True),
        sa.Column('resource_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['resource_id'],
            ['resource.id'],
            name='fk_resource_id',
            ondelete='CASCADE',
        ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine='InnoDB',
    )



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.engine.clients.os import cinder as c_plugin
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

cinder_volume_type_encryption = {
    'heat_template_version': '2015-04-30',
    'resources': {
        'my_encrypted_vol_type': {
            'type': 'OS::Cinder::EncryptedVolumeType',
            'properties': {
                'provider': 'nova.volume.encryptors.luks.LuksEncryptor',
                'control_location': 'front-end',
                'cipher': 'aes-xts-plain64',
                'key_size': '512',
                'volume_type': '01bd581d-33fe-4d6d-bd7b-70ae076d39fb'
            }
        }
    }
}


class CinderEncryptedVolumeTypeTest(common.HeatTestCase):
    def setUp(self):
        super(CinderEncryptedVolumeTypeTest, self).setUp()

        self.ctx = utils.dummy_context()
        self.patchobject(c_plugin.CinderClientPlugin, 'has_extension',
                         return_value=True)
        self.stack = stack.Stack(
            self.ctx, 'cinder_vol_type_encryption_test_stack',
            template.Template(cinder_volume_type_encryption)
        )

        self.my_encrypted_vol_type = self.stack['my_encrypted_vol_type']
        cinder = mock.MagicMock()
        self.cinderclient = mock.MagicMock()
        self.my_encrypted_vol_type.client = cinder
        cinder.return_value = self.cinderclient
        self.volume_encryption_types = (
            self.cinderclient.volume_encryption_types)

    def test_handle_create(self):
        value = mock.MagicMock()
        volume_type_id = '01bd581d-33fe-4d6d-bd7b-70ae076d39fb'
        value.volume_type_id = volume_type_id
        self.volume_encryption_types.create.return_value = value

        with mock.patch.object(self.my_encrypted_vol_type.client_plugin(),
                               'get_volume_type') as mock_get_volume_type:
            mock_get_volume_type.return_value = volume_type_id
            self.my_encrypted_vol_type.handle_create()
            mock_get_volume_type.assert_called_once_with(volume_type_id)

        specs = {
            'control_location': 'front-end',
            'cipher': 'aes-xts-plain64',
            'key_size': 512,
            'provider': 'nova.volume.encryptors.luks.LuksEncryptor'
        }
        self.volume_encryption_types.create.assert_called_once_with(
            volume_type=volume_type_id, specs=specs)
        self.assertEqual(volume_type_id,
                         self.my_encrypted_vol_type.resource_id)

    def test_handle_update(self):
        update_args = {
            'control_location': 'back-end',
            'key_size': 256,
            'cipher': 'aes-cbc-essiv',
            'provider':
                'nova.volume.encryptors.cryptsetup.CryptsetupEncryptor'
        }
        volume_type_id = '01bd581d-33fe-4d6d-bd7b-70ae076d39fb'
        self.my_encrypted_vol_type.resource_id = volume_type_id
        self.my_encrypted_vol_type.handle_update(json_snippet=None,
                                                 tmpl_diff=None,
                                                 prop_diff=update_args)

        self.volume_encryption_types.update.assert_called_once_with(
            volume_type=volume_type_id, specs=update_args)

    def test_get_live_state(self):
        self.my_encrypted_vol_type.resource_id = '1234'
        value = mock.MagicMock()
        value.to_dict.return_value = {
            'volume_type_id': '1122',
            'provider': 'nova.Test',
            'cipher': 'aes-xts-plain64',
            'control_location': 'front-end',
            'key_size': 256
        }
        self.volume_encryption_types.get.return_value = value

        reality = self.my_encrypted_vol_type.get_live_state(
            self.my_encrypted_vol_type.properties)
        expected = {
            'provider': 'nova.Test',
            'cipher': 'aes-xts-plain64',
            'control_location': 'front-end',
            'key_size': 256
        }

        self.assertEqual(expected, reality)

    def test_get_live_state_found_but_deleted(self):
        self.my_encrypted_vol_type.resource_id = '1234'
        value = mock.MagicMock(spec=[])
        self.volume_encryption_types.get.return_value = value

        self.assertRaises(exception.EntityNotFound,
                          self.my_encrypted_vol_type.get_live_state,
                          self.my_encrypted_vol_type.properties)



#
# Copyright (c) 2013 Docker, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import string
from unittest import mock


class APIError(Exception):
    def __init__(self, content, response):
        super(APIError, self).__init__(content)
        self.response = response


errors = mock.MagicMock()
errors.APIError = APIError


class FakeResponse (object):
    def __init__(self, status_code=200, reason='OK'):
        self.status_code = status_code
        self.reason = reason
        self.content = reason


class Client(object):

    def __init__(self, endpoint=None):
        self._endpoint = endpoint
        self._containers = {}
        self.pulled_images = []
        self.container_create = []
        self.container_start = []
        self.version_info = {}

    def _generate_string(self, n=32):
        return ''.join(random.choice(string.ascii_lowercase) for i in range(n))

    def _check_exists(self, container_id):
        if container_id not in self._containers:
            raise APIError(
                '404 Client Error: Not Found ("No such container: '
                '{0}")'.format(container_id),
                FakeResponse(status_code=404,
                             reason='No such container'))

    def _set_running(self, container_id, running):
        self._check_exists(container_id)
        self._containers[container_id] = running

    def inspect_container(self, container_id):
        self._check_exists(container_id)
        info = {
            'Id': container_id,
            'NetworkSettings': {
                'Bridge': 'docker0',
                'Gateway': '172.17.42.1',
                'IPAddress': '172.17.0.3',
                'IPPrefixLen': 16,
                'Ports': {
                    '80/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '1080'}]
                }
            },
            'State': {
                'Running': self._containers[container_id]
            }
        }
        return info

    def logs(self, container_id):
        logs = ['---logs_begin---']
        for i in range(random.randint(1, 20)):
            logs.append(self._generate_string(random.randint(5, 42)))
        logs.append('---logs_end---')
        return '\n'.join(logs)

    def create_container(self, **kwargs):
        self.container_create.append(kwargs)
        container_id = self._generate_string()
        self._containers[container_id] = None
        self._set_running(container_id, False)
        return self.inspect_container(container_id)

    def remove_container(self, container_id, **kwargs):
        self._check_exists(container_id)
        del self._containers[container_id]

    def start(self, container_id, **kwargs):
        self.container_start.append(kwargs)
        self._set_running(container_id, True)

    def stop(self, container_id):
        self._set_running(container_id, False)

    def kill(self, container_id):
        self._set_running(container_id, False)

    def pull(self, image):
        self.pulled_images.append(image)

    def version(self, api_version=True):
        if not self.version_info:
            self.version_info['ApiVersion'] = '1.15'
        return self.version_info

    def set_api_version(self, version):
        self.version_info['ApiVersion'] = version



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import properties
from heat.engine import resource

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class TestResource(resource.Resource):

    PROPERTIES = (
        A, B, C, CA, rA, rB
    ) = (
        'a', 'b', 'c', 'ca', '!a', '!b'
    )

    ATTRIBUTES = (
        A, rA
    ) = (
        'a', '!a'
    )

    properties_schema = {
        A: properties.Schema(
            properties.Schema.STRING,
            _('Fake property a.'),
            default='a',
            update_allowed=True
        ),
        B: properties.Schema(
            properties.Schema.STRING,
            _('Fake property b.'),
            default='b',
            update_allowed=True
        ),
        C: properties.Schema(
            properties.Schema.STRING,
            _('Fake property c.'),
            update_allowed=True,
            default='c'
        ),
        CA: properties.Schema(
            properties.Schema.STRING,
            _('Fake property ca.'),
            update_allowed=True,
            default='ca'
        ),
        rA: properties.Schema(
            properties.Schema.STRING,
            _('Fake property !a.'),
            update_allowed=True,
            default='!a'
        ),
        rB: properties.Schema(
            properties.Schema.STRING,
            _('Fake property !c.'),
            update_allowed=True,
            default='!b'
        ),
    }

    attributes_schema = {
        A: attributes.Schema(
            _('Fake attribute a.'),
            cache_mode=attributes.Schema.CACHE_NONE
        ),
        rA: attributes.Schema(
            _('Fake attribute !a.'),
            cache_mode=attributes.Schema.CACHE_NONE
        ),
    }

    def handle_create(self):
        LOG.info('Creating resource %s with properties %s',
                 self.name, dict(self.properties))
        for prop in self.properties.props.keys():
            self.data_set(prop, self.properties.get(prop), redact=False)

        self.resource_id_set(self.physical_resource_name())

    def handle_update(self, json_snippet=None, tmpl_diff=None, prop_diff=None):
        LOG.info('Updating resource %s with prop_diff %s',
                 self.name, prop_diff)
        for prop in prop_diff:
            if '!' in prop:
                raise resource.UpdateReplace(self.name)
            self.data_set(prop, prop_diff.get(prop), redact=False)

    def _resolve_attribute(self, name):
        if name in self.attributes:
            return self.data().get(name)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class GetRes(object):
    def __init__(self, target_name):
        self.target_name = target_name

    def __repr__(self):
        return 'GetRes(%r)' % self.target_name


class GetAtt(GetRes):
    def __init__(self, target_name, attr):
        super(GetAtt, self).__init__(target_name)
        self.attr = attr

    def __repr__(self):
        return 'GetAtt(%r, %r)' % (self.target_name, self.attr)


class RsrcDef(object):
    def __init__(self, properties, depends_on):
        self.properties = properties
        self.depends_on = depends_on

    def __repr__(self):
        return 'RsrcDef(%r, %r)' % (self.properties, self.depends_on)


class Template(object):

    def __init__(self, resources={}, key=None):
        self.key = key
        self.resources = resources

    def __repr__(self):
        return 'Template(%r)' % self.resources



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutronclient.common import exceptions
from neutronclient.v2_0 import client as neutronclient
from oslo_config import cfg

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine.resources.openstack.neutron import firewall
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


firewall_template = '''
heat_template_version: 2015-04-30
description: Template to test neutron firewall resource
resources:
  firewall:
    type: OS::Neutron::Firewall
    properties:
      name: test-firewall
      firewall_policy_id: policy-id
      admin_state_up: True
      shared: True
      value_specs:
         router_ids:
           - router_1
           - router_2
'''

firewall_policy_template = '''
heat_template_version: 2015-04-30
description: Template to test neutron firewall policy resource
resources:
  firewall_policy:
    type: OS::Neutron::FirewallPolicy
    properties:
      name: test-firewall-policy
      shared: True
      audited: True
      firewall_rules:
        - rule-id-1
        - rule-id-2
'''

firewall_rule_template = '''
heat_template_version: 2015-04-30
description: Template to test neutron firewall rule resource
resources:
  firewall_rule:
    type: OS::Neutron::FirewallRule
    properties:
      name: test-firewall-rule
      shared: True
      protocol: tcp
      action: allow
      enabled: True
      ip_version: 4
'''


class FirewallTest(common.HeatTestCase):

    def setUp(self):
        super(FirewallTest, self).setUp()
        self.mockclient = mock.Mock(spec=neutronclient.Client)
        self.patchobject(neutronclient, 'Client', return_value=self.mockclient)
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)

    def create_firewall(self, value_specs=True):
        snippet = template_format.parse(firewall_template)
        self.mockclient.create_firewall.return_value = {
            'firewall': {'id': '5678'}
        }
        if not value_specs:
            del snippet['resources']['firewall']['properties']['value_specs']

        self.stack = utils.parse_stack(snippet)
        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.fw_props = snippet['resources']['firewall']['properties']
        return firewall.Firewall(
            'firewall', resource_defns['firewall'], self.stack)

    def test_create(self):
        rsrc = self.create_firewall()
        self.mockclient.show_firewall.return_value = {
            'firewall': {'status': 'ACTIVE'}
        }
        scheduler.TaskRunner(rsrc.create)()
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.show_firewall.assert_called_once_with('5678')

    def test_create_failed_error_status(self):
        cfg.CONF.set_override('action_retry_limit', 0)
        rsrc = self.create_firewall()
        self.mockclient.show_firewall.side_effect = [
            {'firewall': {'status': 'PENDING_CREATE'}},
            {'firewall': {'status': 'ERROR'}},
        ]

        error = self.assertRaises(exception.ResourceFailure,
                                  scheduler.TaskRunner(rsrc.create))
        self.assertEqual(
            'ResourceInError: resources.firewall: '
            'Went to status ERROR due to "Error in Firewall"',
            str(error))
        self.assertEqual((rsrc.CREATE, rsrc.FAILED), rsrc.state)

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.show_firewall.assert_called_with('5678')

    def test_create_failed(self):
        self.mockclient.create_firewall.side_effect = (
            exceptions.NeutronClientException())

        snippet = template_format.parse(firewall_template)
        stack = utils.parse_stack(snippet)
        resource_defns = stack.t.resource_definitions(stack)
        rsrc = firewall.Firewall(
            'firewall', resource_defns['firewall'], stack)

        error = self.assertRaises(exception.ResourceFailure,
                                  scheduler.TaskRunner(rsrc.create))
        self.assertEqual(
            'NeutronClientException: resources.firewall: '
            'An unknown exception occurred.',
            str(error))
        self.assertEqual((rsrc.CREATE, rsrc.FAILED), rsrc.state)

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })

    def test_delete(self):
        rsrc = self.create_firewall()
        self.mockclient.show_firewall.side_effect = [
            {'firewall': {'status': 'ACTIVE'}},
            exceptions.NeutronClientException(status_code=404),
        ]
        self.mockclient.delete_firewall.return_value = None

        scheduler.TaskRunner(rsrc.create)()
        scheduler.TaskRunner(rsrc.delete)()
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.delete_firewall.assert_called_once_with('5678')
        self.mockclient.show_firewall.assert_called_with('5678')

    def test_delete_already_gone(self):
        rsrc = self.create_firewall()
        self.mockclient.show_firewall.return_value = {
            'firewall': {'status': 'ACTIVE'}
        }
        self.mockclient.delete_firewall.side_effect = (
            exceptions.NeutronClientException(status_code=404))

        scheduler.TaskRunner(rsrc.create)()
        scheduler.TaskRunner(rsrc.delete)()
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.delete_firewall.assert_called_once_with('5678')
        self.mockclient.show_firewall.assert_called_once_with('5678')

    def test_delete_failed(self):
        rsrc = self.create_firewall()
        self.mockclient.show_firewall.return_value = {
            'firewall': {'status': 'ACTIVE'}
        }
        self.mockclient.delete_firewall.side_effect = (
            exceptions.NeutronClientException(status_code=400))

        scheduler.TaskRunner(rsrc.create)()
        error = self.assertRaises(exception.ResourceFailure,
                                  scheduler.TaskRunner(rsrc.delete))
        self.assertEqual(
            'NeutronClientException: resources.firewall: '
            'An unknown exception occurred.',
            str(error))
        self.assertEqual((rsrc.DELETE, rsrc.FAILED), rsrc.state)

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.delete_firewall.assert_called_once_with('5678')
        self.mockclient.show_firewall.assert_called_once_with('5678')

    def test_attribute(self):
        rsrc = self.create_firewall()
        self.mockclient.show_firewall.return_value = {
            'firewall': {
                'status': 'ACTIVE',
                'admin_state_up': True,
                'firewall_policy_id': 'policy-id',
                'shared': True,
            }
        }

        scheduler.TaskRunner(rsrc.create)()
        self.assertIs(True, rsrc.FnGetAtt('admin_state_up'))
        self.assertEqual('This attribute is currently unsupported in neutron '
                         'firewall resource.', rsrc.FnGetAtt('shared'))
        self.assertEqual('policy-id', rsrc.FnGetAtt('firewall_policy_id'))

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.show_firewall.assert_called_with('5678')

    def test_attribute_failed(self):
        rsrc = self.create_firewall()
        self.mockclient.show_firewall.return_value = {
            'firewall': {'status': 'ACTIVE'}
        }

        scheduler.TaskRunner(rsrc.create)()
        error = self.assertRaises(exception.InvalidTemplateAttribute,
                                  rsrc.FnGetAtt, 'subnet_id')
        self.assertEqual(
            'The Referenced Attribute (firewall subnet_id) is '
            'incorrect.', str(error))

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.show_firewall.assert_called_once_with('5678')

    def test_update(self):
        rsrc = self.create_firewall()
        self.mockclient.show_firewall.return_value = {
            'firewall': {'status': 'ACTIVE'}
        }
        self.mockclient.update_firewall.return_value = None

        scheduler.TaskRunner(rsrc.create)()

        props = self.fw_props.copy()
        props['admin_state_up'] = False
        update_template = rsrc.t.freeze(properties=props)
        scheduler.TaskRunner(rsrc.update, update_template)()

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.show_firewall.assert_called_once_with('5678')
        self.mockclient.update_firewall.assert_called_once_with(
            '5678', {'firewall': {'admin_state_up': False}})

    def test_update_with_value_specs(self):
        rsrc = self.create_firewall(value_specs=False)
        self.mockclient.show_firewall.return_value = {
            'firewall': {'status': 'ACTIVE'}
        }

        scheduler.TaskRunner(rsrc.create)()
        prop_diff = {
            'value_specs': {
                'router_ids': ['router_1', 'router_2']
            }
        }
        update_snippet = rsrc_defn.ResourceDefinition(rsrc.name,
                                                      rsrc.type(),
                                                      prop_diff)
        rsrc.handle_update(update_snippet, {}, prop_diff)

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.show_firewall.assert_called_once_with('5678')
        self.mockclient.update_firewall.assert_called_once_with(
            '5678', {'firewall': {'router_ids': ['router_1',
                                                 'router_2']}})

    def test_get_live_state(self):
        rsrc = self.create_firewall(value_specs=True)
        self.mockclient.show_firewall.return_value = {
            'firewall': {
                'status': 'ACTIVE',
                'router_ids': ['router_1', 'router_2'],
                'name': 'firewall-firewall-pwakkqdrcl7z',
                'admin_state_up': True,
                'tenant_id': 'df49ea64e87c43a792a510698364f03e',
                'firewall_policy_id': '680eb26d-3eea-40be-b484-1476e4c7c1b3',
                'id': '11425cd4-41b6-4fd4-97aa-17629c63de61',
                'description': ''
            }
        }

        scheduler.TaskRunner(rsrc.create)()

        reality = rsrc.get_live_state(rsrc.properties)
        expected = {
            'value_specs': {
                'router_ids': ['router_1', 'router_2']
            },
            'name': 'firewall-firewall-pwakkqdrcl7z',
            'admin_state_up': True,
            'firewall_policy_id': '680eb26d-3eea-40be-b484-1476e4c7c1b3',
            'description': ''
        }

        self.assertEqual(expected, reality)

        self.mockclient.create_firewall.assert_called_once_with({
            'firewall': {
                'name': 'test-firewall', 'admin_state_up': True,
                'router_ids': ['router_1', 'router_2'],
                'firewall_policy_id': 'policy-id', 'shared': True
            }
        })
        self.mockclient.show_firewall.assert_called_with('5678')


class FirewallPolicyTest(common.HeatTestCase):

    def setUp(self):
        super(FirewallPolicyTest, self).setUp()
        self.mockclient = mock.Mock(spec=neutronclient.Client)
        self.patchobject(neutronclient, 'Client', return_value=self.mockclient)
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)

    def create_firewall_policy(self):
        self.mockclient.create_firewall_policy.return_value = {
            'firewall_policy': {'id': '5678'}
        }

        snippet = template_format.parse(firewall_policy_template)
        self.stack = utils.parse_stack(snippet)
        self.tmpl = snippet
        resource_defns = self.stack.t.resource_definitions(self.stack)
        return firewall.FirewallPolicy(
            'firewall_policy', resource_defns['firewall_policy'], self.stack)

    def test_create(self):
        rsrc = self.create_firewall_policy()

        scheduler.TaskRunner(rsrc.create)()
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall_policy.assert_called_once_with({
            'firewall_policy': {
                'name': 'test-firewall-policy', 'shared': True,
                'audited': True, 'firewall_rules': ['rule-id-1', 'rule-id-2']
            }
        })

    def test_create_failed(self):
        self.mockclient.create_firewall_policy.side_effect = (
            exceptions.NeutronClientException())

        snippet = template_format.parse(firewall_policy_template)
        stack = utils.parse_stack(snippet)
        resource_defns = stack.t.resource_definitions(stack)
        rsrc = firewall.FirewallPolicy(
            'firewall_policy', resource_defns['firewall_policy'], stack)

        error = self.assertRaises(exception.ResourceFailure,
                                  scheduler.TaskRunner(rsrc.create))
        self.assertEqual(
            'NeutronClientException: resources.firewall_policy: '
            'An unknown exception occurred.',
            str(error))
        self.assertEqual((rsrc.CREATE, rsrc.FAILED), rsrc.state)

        self.mockclient.create_firewall_policy.assert_called_once_with({
            'firewall_policy': {
                'name': 'test-firewall-policy', 'shared': True,
                'audited': True, 'firewall_rules': ['rule-id-1', 'rule-id-2']
            }
        })

    def test_delete(self):
        rsrc = self.create_firewall_policy()
        self.mockclient.delete_firewall_policy.return_value = None
        self.mockclient.show_firewall_policy.side_effect = (
            exceptions.NeutronClientException(status_code=404))

        scheduler.TaskRunner(rsrc.create)()
        scheduler.TaskRunner(rsrc.delete)()
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall_policy.assert_called_once_with({
            'firewall_policy': {
                'name': 'test-firewall-policy', 'shared': True,
                'audited': True, 'firewall_rules': ['rule-id-1', 'rule-id-2']
            }
        })
        self.mockclient.delete_firewall_policy.assert_called_once_with('5678')
        self.mockclient.show_firewall_policy.assert_called_once_with('5678')

    def test_delete_already_gone(self):
        rsrc = self.create_firewall_policy()
        self.mockclient.delete_firewall_policy.side_effect = (
            exceptions.NeutronClientException(status_code=404))

        scheduler.TaskRunner(rsrc.create)()
        scheduler.TaskRunner(rsrc.delete)()
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall_policy.assert_called_once_with({
            'firewall_policy': {
                'name': 'test-firewall-policy', 'shared': True,
                'audited': True, 'firewall_rules': ['rule-id-1', 'rule-id-2']
            }
        })
        self.mockclient.delete_firewall_policy.assert_called_once_with('5678')
        self.mockclient.show_firewall_policy.assert_not_called()

    def test_delete_failed(self):
        rsrc = self.create_firewall_policy()
        self.mockclient.delete_firewall_policy.side_effect = (
            exceptions.NeutronClientException(status_code=400))

        scheduler.TaskRunner(rsrc.create)()
        error = self.assertRaises(exception.ResourceFailure,
                                  scheduler.TaskRunner(rsrc.delete))
        self.assertEqual(
            'NeutronClientException: resources.firewall_policy: '
            'An unknown exception occurred.',
            str(error))
        self.assertEqual((rsrc.DELETE, rsrc.FAILED), rsrc.state)

        self.mockclient.create_firewall_policy.assert_called_once_with({
            'firewall_policy': {
                'name': 'test-firewall-policy', 'shared': True,
                'audited': True, 'firewall_rules': ['rule-id-1', 'rule-id-2']
            }
        })
        self.mockclient.delete_firewall_policy.assert_called_once_with('5678')
        self.mockclient.show_firewall_policy.assert_not_called()

    def test_attribute(self):
        rsrc = self.create_firewall_policy()
        self.mockclient.show_firewall_policy.return_value = {
            'firewall_policy': {'audited': True, 'shared': True}
        }

        scheduler.TaskRunner(rsrc.create)()
        self.assertIs(True, rsrc.FnGetAtt('audited'))
        self.assertIs(True, rsrc.FnGetAtt('shared'))

        self.mockclient.create_firewall_policy.assert_called_once_with({
            'firewall_policy': {
                'name': 'test-firewall-policy', 'shared': True,
                'audited': True, 'firewall_rules': ['rule-id-1', 'rule-id-2']
            }
        })
        self.mockclient.show_firewall_policy.assert_called_with('5678')

    def test_attribute_failed(self):
        rsrc = self.create_firewall_policy()

        scheduler.TaskRunner(rsrc.create)()
        error = self.assertRaises(exception.InvalidTemplateAttribute,
                                  rsrc.FnGetAtt, 'subnet_id')
        self.assertEqual(
            'The Referenced Attribute (firewall_policy subnet_id) is '
            'incorrect.', str(error))

        self.mockclient.create_firewall_policy.assert_called_once_with({
            'firewall_policy': {
                'name': 'test-firewall-policy', 'shared': True,
                'audited': True, 'firewall_rules': ['rule-id-1', 'rule-id-2']
            }
        })
        self.mockclient.show_firewall_policy.assert_not_called()

    def test_update(self):
        rsrc = self.create_firewall_policy()
        self.mockclient.update_firewall_policy.return_value = None

        scheduler.TaskRunner(rsrc.create)()

        props = self.tmpl['resources']['firewall_policy']['properties'].copy()
        props['firewall_rules'] = ['3', '4']
        update_template = rsrc.t.freeze(properties=props)
        scheduler.TaskRunner(rsrc.update, update_template)()

        self.mockclient.create_firewall_policy.assert_called_once_with({
            'firewall_policy': {
                'name': 'test-firewall-policy', 'shared': True,
                'audited': True, 'firewall_rules': ['rule-id-1', 'rule-id-2']
            }
        })
        self.mockclient.update_firewall_policy.assert_called_once_with(
            '5678', {'firewall_policy': {'firewall_rules': ['3', '4']}})


class FirewallRuleTest(common.HeatTestCase):

    def setUp(self):
        super(FirewallRuleTest, self).setUp()
        self.mockclient = mock.Mock(spec=neutronclient.Client)
        self.patchobject(neutronclient, 'Client', return_value=self.mockclient)
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)

    def create_firewall_rule(self):
        self.mockclient.create_firewall_rule.return_value = {
            'firewall_rule': {'id': '5678'}
        }

        snippet = template_format.parse(firewall_rule_template)
        self.stack = utils.parse_stack(snippet)
        self.tmpl = snippet
        resource_defns = self.stack.t.resource_definitions(self.stack)
        return firewall.FirewallRule(
            'firewall_rule', resource_defns['firewall_rule'], self.stack)

    def test_create(self):
        rsrc = self.create_firewall_rule()

        scheduler.TaskRunner(rsrc.create)()
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })

    def test_validate_failed_with_string_None_protocol(self):
        snippet = template_format.parse(firewall_rule_template)
        stack = utils.parse_stack(snippet)
        rsrc = stack['firewall_rule']
        props = dict(rsrc.properties)
        props['protocol'] = 'None'
        rsrc.t = rsrc.t.freeze(properties=props)
        rsrc.reparse()
        self.assertRaises(exception.StackValidationFailed, rsrc.validate)

    def test_create_with_protocol_any(self):
        self.mockclient.create_firewall_rule.return_value = {
            'firewall_rule': {'id': '5678'}
        }

        snippet = template_format.parse(firewall_rule_template)
        snippet['resources']['firewall_rule']['properties']['protocol'] = 'any'
        stack = utils.parse_stack(snippet)
        rsrc = stack['firewall_rule']

        scheduler.TaskRunner(rsrc.create)()
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': None, 'enabled': True,
                'ip_version': "4"
            }
        })

    def test_create_failed(self):
        self.mockclient.create_firewall_rule.side_effect = (
            exceptions.NeutronClientException())

        snippet = template_format.parse(firewall_rule_template)
        stack = utils.parse_stack(snippet)
        resource_defns = stack.t.resource_definitions(stack)
        rsrc = firewall.FirewallRule(
            'firewall_rule', resource_defns['firewall_rule'], stack)

        error = self.assertRaises(exception.ResourceFailure,
                                  scheduler.TaskRunner(rsrc.create))
        self.assertEqual(
            'NeutronClientException: resources.firewall_rule: '
            'An unknown exception occurred.',
            str(error))
        self.assertEqual((rsrc.CREATE, rsrc.FAILED), rsrc.state)

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })

    def test_delete(self):
        rsrc = self.create_firewall_rule()
        self.mockclient.delete_firewall_rule.return_value = None
        self.mockclient.show_firewall_rule.side_effect = (
            exceptions.NeutronClientException(status_code=404))

        scheduler.TaskRunner(rsrc.create)()
        scheduler.TaskRunner(rsrc.delete)()
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })
        self.mockclient.delete_firewall_rule.assert_called_once_with('5678')
        self.mockclient.show_firewall_rule.assert_called_once_with('5678')

    def test_delete_already_gone(self):
        rsrc = self.create_firewall_rule()
        self.mockclient.delete_firewall_rule.side_effect = (
            exceptions.NeutronClientException(status_code=404))

        scheduler.TaskRunner(rsrc.create)()
        scheduler.TaskRunner(rsrc.delete)()
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })
        self.mockclient.delete_firewall_rule.assert_called_once_with('5678')
        self.mockclient.show_firewall_rule.assert_not_called()

    def test_delete_failed(self):
        rsrc = self.create_firewall_rule()
        self.mockclient.delete_firewall_rule.side_effect = (
            exceptions.NeutronClientException(status_code=400))

        scheduler.TaskRunner(rsrc.create)()
        error = self.assertRaises(exception.ResourceFailure,
                                  scheduler.TaskRunner(rsrc.delete))
        self.assertEqual(
            'NeutronClientException: resources.firewall_rule: '
            'An unknown exception occurred.',
            str(error))
        self.assertEqual((rsrc.DELETE, rsrc.FAILED), rsrc.state)

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })
        self.mockclient.delete_firewall_rule.assert_called_once_with('5678')
        self.mockclient.show_firewall_rule.assert_not_called()

    def test_attribute(self):
        rsrc = self.create_firewall_rule()
        self.mockclient.show_firewall_rule.return_value = {
            'firewall_rule': {'protocol': 'tcp', 'shared': True}
        }

        scheduler.TaskRunner(rsrc.create)()
        self.assertEqual('tcp', rsrc.FnGetAtt('protocol'))
        self.assertIs(True, rsrc.FnGetAtt('shared'))

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })
        self.mockclient.show_firewall_rule.assert_called_with('5678')

    def test_attribute_failed(self):
        rsrc = self.create_firewall_rule()

        scheduler.TaskRunner(rsrc.create)()
        error = self.assertRaises(exception.InvalidTemplateAttribute,
                                  rsrc.FnGetAtt, 'subnet_id')
        self.assertEqual(
            'The Referenced Attribute (firewall_rule subnet_id) is '
            'incorrect.', str(error))

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })
        self.mockclient.show_firewall_rule.assert_not_called()

    def test_update(self):
        rsrc = self.create_firewall_rule()
        self.mockclient.update_firewall_rule.return_value = None

        scheduler.TaskRunner(rsrc.create)()

        props = self.tmpl['resources']['firewall_rule']['properties'].copy()
        props['protocol'] = 'icmp'
        update_template = rsrc.t.freeze(properties=props)
        scheduler.TaskRunner(rsrc.update, update_template)()

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })
        self.mockclient.update_firewall_rule.assert_called_once_with(
            '5678', {'firewall_rule': {'protocol': 'icmp'}})

    def test_update_protocol_to_any(self):
        rsrc = self.create_firewall_rule()
        self.mockclient.update_firewall_rule.return_value = None

        scheduler.TaskRunner(rsrc.create)()
        # update to 'any' protocol
        props = self.tmpl['resources']['firewall_rule']['properties'].copy()
        props['protocol'] = 'any'
        update_template = rsrc.t.freeze(properties=props)
        scheduler.TaskRunner(rsrc.update, update_template)()

        self.mockclient.create_firewall_rule.assert_called_once_with({
            'firewall_rule': {
                'name': 'test-firewall-rule', 'shared': True,
                'action': 'allow', 'protocol': 'tcp', 'enabled': True,
                'ip_version': "4"
            }
        })
        self.mockclient.update_firewall_rule.assert_called_once_with(
            '5678', {'firewall_rule': {'protocol': None}})



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from zaqarclient.queues.v2 import client as zaqarclient
from zaqarclient.transport import errors as zaqar_errors

from heat.common.i18n import _
from heat.engine.clients import client_plugin
from heat.engine import constraints

LOG = logging.getLogger(__name__)

CLIENT_NAME = 'zaqar'


class ZaqarClientPlugin(client_plugin.ClientPlugin):

    exceptions_module = zaqar_errors

    service_types = [MESSAGING] = ['messaging']

    DEFAULT_TTL = 3600

    def _create(self):
        return zaqarclient.Client(version=2,
                                  session=self.context.keystone_session)

    def create_for_tenant(self, tenant_id, token):
        con = self.context
        if token is None:
            LOG.error("Zaqar connection failed, no auth_token!")
            return None

        opts = {
            'os_auth_token': token,
            'os_auth_url': con.auth_url,
            'os_project_id': tenant_id,
            'os_service_type': self.MESSAGING,
        }
        auth_opts = {'backend': 'keystone',
                     'options': opts}
        conf = {'auth_opts': auth_opts}
        endpoint = self.url_for(service_type=self.MESSAGING)
        return zaqarclient.Client(url=endpoint, conf=conf, version=2)

    def create_from_signed_url(self, project_id, paths, expires, methods,
                               signature):
        opts = {
            'paths': paths,
            'expires': expires,
            'methods': methods,
            'signature': signature,
            'os_project_id': project_id,
        }
        auth_opts = {'backend': 'signed-url',
                     'options': opts}
        conf = {'auth_opts': auth_opts}
        endpoint = self.url_for(service_type=self.MESSAGING)
        return zaqarclient.Client(url=endpoint, conf=conf, version=2)

    def is_not_found(self, ex):
        return isinstance(ex, zaqar_errors.ResourceNotFound)

    def get_queue(self, queue_name):
        if not isinstance(queue_name, str):
            raise TypeError(_('Queue name must be a string'))
        if not (0 < len(queue_name) <= 64):
            raise ValueError(_('Queue name length must be 1-64'))
        # Queues are created automatically starting with v1.1 of the Zaqar API,
        # so any queue name is valid for the purposes of constraint validation.
        return queue_name


class ZaqarEventSink(object):

    def __init__(self, target, ttl=None):
        self._target = target
        self._ttl = ttl

    def consume(self, context, event):
        zaqar_plugin = context.clients.client_plugin('zaqar')
        zaqar = zaqar_plugin.client()
        queue = zaqar.queue(self._target, auto_create=False)
        ttl = self._ttl if self._ttl is not None else zaqar_plugin.DEFAULT_TTL
        queue.post({'body': event, 'ttl': ttl})


class QueueConstraint(constraints.BaseCustomConstraint):
    resource_client_name = CLIENT_NAME
    resource_getter_name = 'get_queue'



#
# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import webob
from webob import exc

from heat.common import auth_url
from heat.tests import common


class FakeApp(object):
    """This represents a WSGI app protected by our auth middleware."""

    def __call__(self, environ, start_response):
        """Assert that headers are correctly set up when finally called."""
        resp = webob.Response()
        resp.body = 'SUCCESS'.encode('latin-1')
        return resp(environ, start_response)


class AuthUrlFilterTest(common.HeatTestCase):

    def setUp(self):
        super(AuthUrlFilterTest, self).setUp()
        self.app = FakeApp()
        self.config = {'auth_uri': 'foobar'}
        self.middleware = auth_url.AuthUrlFilter(self.app, self.config)

    @mock.patch.object(auth_url.cfg, 'CONF')
    def test_adds_default_auth_url_from_clients_keystone(self, mock_cfg):
        self.config = {}
        mock_cfg.clients_keystone.auth_uri = 'foobar'
        mock_cfg.keystone_authtoken.www_authenticate_uri = 'should-be-ignored'
        mock_cfg.auth_password.multi_cloud = False
        with mock.patch('keystoneauth1.discover.Discover') as discover:
            class MockDiscover(object):
                def url_for(self, endpoint):
                    return 'foobar/v3'
            discover.return_value = MockDiscover()
            self.middleware = auth_url.AuthUrlFilter(self.app, self.config)
            req = webob.Request.blank('/tenant_id/')
            self.middleware(req)
            self.assertIn('X-Auth-Url', req.headers)
            self.assertEqual('foobar/v3', req.headers['X-Auth-Url'])

    @mock.patch.object(auth_url.cfg, 'CONF')
    def test_adds_default_auth_url_from_keystone_authtoken(self, mock_cfg):
        self.config = {}
        mock_cfg.clients_keystone.auth_uri = ''
        mock_cfg.keystone_authtoken.www_authenticate_uri = 'foobar'
        mock_cfg.auth_password.multi_cloud = False
        self.middleware = auth_url.AuthUrlFilter(self.app, self.config)
        req = webob.Request.blank('/tenant_id/')
        self.middleware(req)
        self.assertIn('X-Auth-Url', req.headers)
        self.assertEqual('foobar', req.headers['X-Auth-Url'])

    def test_overwrites_auth_url_from_headers_with_local_config(self):
        req = webob.Request.blank('/tenant_id/')
        req.headers['X-Auth-Url'] = 'should_be_overwritten'
        self.middleware(req)
        self.assertEqual('foobar', req.headers['X-Auth-Url'])

    def test_reads_auth_url_from_local_config(self):
        req = webob.Request.blank('/tenant_id/')
        self.middleware(req)
        self.assertIn('X-Auth-Url', req.headers)
        self.assertEqual('foobar', req.headers['X-Auth-Url'])

    @mock.patch.object(auth_url.AuthUrlFilter, '_validate_auth_url')
    @mock.patch.object(auth_url.cfg, 'CONF')
    def test_multicloud_reads_auth_url_from_headers(self, mock_cfg, mock_val):
        mock_cfg.auth_password.multi_cloud = True
        mock_val.return_value = True
        req = webob.Request.blank('/tenant_id/')
        req.headers['X-Auth-Url'] = 'overwrites config'
        self.middleware(req)
        self.assertIn('X-Auth-Url', req.headers)
        self.assertEqual('overwrites config', req.headers['X-Auth-Url'])

    @mock.patch.object(auth_url.AuthUrlFilter, '_validate_auth_url')
    @mock.patch.object(auth_url.cfg, 'CONF')
    def test_multicloud_validates_auth_url(self, mock_cfg, mock_validate):
        mock_cfg.auth_password.multi_cloud = True
        req = webob.Request.blank('/tenant_id/')

        self.middleware(req)
        self.assertTrue(mock_validate.called)

    def test_validate_auth_url_with_missing_url(self):
        self.assertRaises(exc.HTTPBadRequest,
                          self.middleware._validate_auth_url,
                          auth_url='')

        self.assertRaises(exc.HTTPBadRequest,
                          self.middleware._validate_auth_url,
                          auth_url=None)

    @mock.patch.object(auth_url.cfg, 'CONF')
    def test_validate_auth_url_with_url_not_allowed(self, mock_cfg):
        mock_cfg.auth_password.allowed_auth_uris = ['foobar']

        self.assertRaises(exc.HTTPUnauthorized,
                          self.middleware._validate_auth_url,
                          auth_url='not foobar')

    @mock.patch.object(auth_url.cfg, 'CONF')
    def test_validate_auth_url_with_valid_url(self, mock_cfg):
        mock_cfg.auth_password.allowed_auth_uris = ['foobar']

        self.assertTrue(self.middleware._validate_auth_url('foobar'))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.engine.clients.os import cinder as c_plugin
from heat.engine.resources.openstack.cinder import qos_specs
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

QOS_SPECS_TEMPLATE = {
    'heat_template_version': '2015-10-15',
    'description': 'Cinder QoS specs creation example',
    'resources': {
        'my_qos_specs': {
            'type': 'OS::Cinder::QoSSpecs',
            'properties': {
                'name': 'foobar',
                'specs': {"foo": "bar", "foo1": "bar1"}
            }
        }
    }
}

QOS_ASSOCIATE_TEMPLATE = {
    'heat_template_version': '2015-10-15',
    'description': 'Cinder QoS specs association example',
    'resources': {
        'my_qos_associate': {
            'type': 'OS::Cinder::QoSAssociation',
            'properties': {
                'volume_types': ['ceph', 'lvm'],
                'qos_specs': 'foobar'
            }
        }
    }
}


class QoSSpecsTest(common.HeatTestCase):

    def setUp(self):
        super(QoSSpecsTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.patchobject(c_plugin.CinderClientPlugin, 'has_extension',
                         return_value=True)
        self.stack = stack.Stack(
            self.ctx, 'cinder_qos_spec_test_stack',
            template.Template(QOS_SPECS_TEMPLATE)
        )
        self.my_qos_specs = self.stack['my_qos_specs']
        cinder_client = mock.MagicMock()
        self.cinderclient = mock.MagicMock()
        self.my_qos_specs.client = cinder_client
        cinder_client.return_value = self.cinderclient
        self.qos_specs = self.cinderclient.qos_specs
        self.value = mock.MagicMock()
        self.value.id = '927202df-1afb-497f-8368-9c2d2f26e5db'
        self.value.name = 'foobar'
        self.value.specs = {'foo': 'bar', 'foo1': 'bar1'}
        self.qos_specs.create.return_value = self.value

    def test_resource_mapping(self):
        mapping = qos_specs.resource_mapping()
        self.assertEqual(2, len(mapping))
        self.assertEqual(qos_specs.QoSSpecs,
                         mapping['OS::Cinder::QoSSpecs'])
        self.assertIsInstance(self.my_qos_specs,
                              qos_specs.QoSSpecs)

    def _set_up_qos_specs_environment(self):
        self.qos_specs.create.return_value = self.value
        self.my_qos_specs.handle_create()

    def test_qos_specs_handle_create_specs(self):
        self._set_up_qos_specs_environment()
        self.assertEqual(1, self.qos_specs.create.call_count)
        self.assertEqual(self.value.id, self.my_qos_specs.resource_id)

    def test_qos_specs_handle_update_specs(self):
        self._set_up_qos_specs_environment()
        resource_id = self.my_qos_specs.resource_id
        prop_diff = {'specs': {'foo': 'bar', 'bar': 'bar'}}
        set_expected = {'bar': 'bar'}
        unset_expected = ['foo1']

        self.my_qos_specs.handle_update(
            json_snippet=None, tmpl_diff=None, prop_diff=prop_diff
        )
        self.qos_specs.set_keys.assert_called_once_with(
            resource_id,
            set_expected
        )
        self.qos_specs.unset_keys.assert_called_once_with(
            resource_id,
            unset_expected
        )

    def test_qos_specs_handle_delete_specs(self):
        self._set_up_qos_specs_environment()
        resource_id = self.my_qos_specs.resource_id
        self.my_qos_specs.handle_delete()
        self.qos_specs.disassociate_all.assert_called_once_with(resource_id)


class QoSAssociationTest(common.HeatTestCase):

    def setUp(self):
        super(QoSAssociationTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.qos_specs_id = 'foobar'
        self.patchobject(c_plugin.CinderClientPlugin, 'has_extension',
                         return_value=True)
        self.patchobject(c_plugin.CinderClientPlugin, 'get_qos_specs',
                         return_value=self.qos_specs_id)
        self.stack = stack.Stack(
            self.ctx, 'cinder_qos_associate_test_stack',
            template.Template(QOS_ASSOCIATE_TEMPLATE)
        )
        self.my_qos_associate = self.stack['my_qos_associate']
        cinder_client = mock.MagicMock()
        self.cinderclient = mock.MagicMock()
        self.my_qos_associate.client = cinder_client
        cinder_client.return_value = self.cinderclient
        self.qos_specs = self.cinderclient.qos_specs
        self.stub_QoSSpecsConstraint_validate()
        self.stub_VolumeTypeConstraint_validate()

        self.vt_ceph = 'ceph'
        self.vt_lvm = 'lvm'
        self.vt_new_ceph = 'new_ceph'

    def test_resource_mapping(self):
        mapping = qos_specs.resource_mapping()
        self.assertEqual(2, len(mapping))
        self.assertEqual(qos_specs.QoSAssociation,
                         mapping['OS::Cinder::QoSAssociation'])
        self.assertIsInstance(self.my_qos_associate,
                              qos_specs.QoSAssociation)

    def _set_up_qos_associate_environment(self):
        self.my_qos_associate.handle_create()

    def test_qos_associate_handle_create(self):
        self.patchobject(c_plugin.CinderClientPlugin, 'get_volume_type',
                         side_effect=[self.vt_ceph, self.vt_lvm])
        self._set_up_qos_associate_environment()
        self.cinderclient.qos_specs.associate.assert_any_call(
            self.qos_specs_id,
            self.vt_ceph
        )
        self.qos_specs.associate.assert_any_call(
            self.qos_specs_id,
            self.vt_lvm
        )

    def test_qos_associate_handle_update(self):
        self.patchobject(c_plugin.CinderClientPlugin, 'get_volume_type',
                         side_effect=[self.vt_lvm, self.vt_ceph,
                                      self.vt_new_ceph,
                                      self.vt_ceph])
        self._set_up_qos_associate_environment()
        prop_diff = {'volume_types': [self.vt_lvm, self.vt_new_ceph]}
        self.my_qos_associate.handle_update(
            json_snippet=None, tmpl_diff=None, prop_diff=prop_diff
        )
        self.qos_specs.associate.assert_any_call(
            self.qos_specs_id,
            self.vt_new_ceph
        )
        self.qos_specs.disassociate.assert_any_call(
            self.qos_specs_id,
            self.vt_ceph
        )

    def test_qos_associate_handle_delete_specs(self):
        self.patchobject(c_plugin.CinderClientPlugin, 'get_volume_type',
                         side_effect=[self.vt_ceph, self.vt_lvm,
                                      self.vt_ceph, self.vt_lvm])
        self._set_up_qos_associate_environment()
        self.my_qos_associate.handle_delete()
        self.qos_specs.disassociate.assert_any_call(
            self.qos_specs_id,
            self.vt_ceph
        )
        self.qos_specs.disassociate.assert_any_call(
            self.qos_specs_id,
            self.vt_lvm
        )



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from heat.engine.clients.os.keystone import fake_keystoneclient as fake_ks
from heat.engine import resource
from heat.engine.resources.openstack.keystone import role
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

keystone_role_template = {
    'heat_template_version': '2013-05-23',
    'resources': {
        'test_role': {
            'type': 'OS::Keystone::Role',
            'properties': {
                'name': 'test_role_1'
            }
        }
    }
}


class KeystoneRoleTest(common.HeatTestCase):
    def setUp(self):
        super(KeystoneRoleTest, self).setUp()

        self.ctx = utils.dummy_context()
        # Mock client
        self.keystoneclient = mock.Mock()
        self.patchobject(resource.Resource, 'client',
                         return_value=fake_ks.FakeKeystoneClient(
                             client=self.keystoneclient))
        self.roles = self.keystoneclient.roles

    def _get_rsrc(self, domain='default', without_name=False):
        t = template.Template(keystone_role_template)
        tmpl = copy.deepcopy(t)
        tmpl['resources']['test_role']['Properties']['domain'] = domain
        if without_name:
            tmpl['resources']['test_role']['Properties'].pop('name')
        test_stack = stack.Stack(self.ctx, 'test_keystone_role', tmpl)
        test_role = test_stack['test_role']
        return test_role

    def _get_mock_role(self, domain='default'):
        value = mock.MagicMock()
        role_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        domain_id = domain
        value.id = role_id
        value.domain_id = domain_id
        return value

    def _test_handle_create(self, domain='default'):
        test_role = self._get_rsrc(domain)
        mock_role = self._get_mock_role(domain)
        self.roles.create.return_value = mock_role

        # validate the properties
        self.assertEqual('test_role_1',
                         test_role.properties.get(role.KeystoneRole.NAME))
        self.assertEqual(domain,
                         test_role.properties.get(role.KeystoneRole.DOMAIN))

        test_role.handle_create()

        # validate role creation with given name
        self.roles.create.assert_called_once_with(name='test_role_1',
                                                  domain=domain)

        # validate physical resource id
        self.assertEqual(mock_role.id, test_role.resource_id)

    def test_role_handle_create(self):
        self._test_handle_create()

    def test_role_handle_create_with_domain(self):
        self._test_handle_create(domain='d_test')

    def test_role_handle_create_default_name(self):
        # reset the NAME value to None, to make sure role is
        # created with physical_resource_name
        test_role = self._get_rsrc(without_name=True)
        test_role.physical_resource_name = mock.Mock(
            return_value='phy_role_name')
        test_role.handle_create()

        # validate role creation with default name
        self.roles.create.assert_called_once_with(name='phy_role_name',
                                                  domain='default')

    def test_role_handle_update(self):
        test_role = self._get_rsrc()
        test_role.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        # update the name property
        prop_diff = {role.KeystoneRole.NAME: 'test_role_1_updated'}

        test_role.handle_update(json_snippet=None,
                                tmpl_diff=None,
                                prop_diff=prop_diff)

        self.roles.update.assert_called_once_with(
            role=test_role.resource_id,
            name=prop_diff[role.KeystoneRole.NAME]
        )

    def test_show_resource(self):
        role = mock.Mock()
        role.to_dict.return_value = {'attr': 'val'}
        self.roles.get.return_value = role
        test_role = self._get_rsrc()
        res = test_role._show_resource()
        self.assertEqual({'attr': 'val'}, res)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Client side of the heat worker RPC API."""

from heat.common import messaging
from heat.rpc import worker_api


class WorkerClient(object):
    """Client side of the heat worker RPC API.

    API version history::

        1.0 - Initial version.
        1.1 - Added check_resource.
        1.2 - Add adopt data argument to check_resource.
        1.3 - Added cancel_check_resource API.
        1.4 - Add converge argument to check_resource
    """

    BASE_RPC_API_VERSION = '1.0'

    def __init__(self):
        self._client = messaging.get_rpc_client(
            topic=worker_api.TOPIC,
            version=self.BASE_RPC_API_VERSION)

    @staticmethod
    def make_msg(method, **kwargs):
        return method, kwargs

    def cast(self, ctxt, msg, version=None):
        method, kwargs = msg
        if version is not None:
            client = self._client.prepare(version=version)
        else:
            client = self._client
        client.cast(ctxt, method, **kwargs)

    def check_resource(self, ctxt, resource_id,
                       current_traversal, data, is_update, adopt_stack_data,
                       converge=False):
        self.cast(ctxt,
                  self.make_msg(
                      'check_resource', resource_id=resource_id,
                      current_traversal=current_traversal, data=data,
                      is_update=is_update, adopt_stack_data=adopt_stack_data,
                      converge=converge
                  ),
                  version='1.4')

    def cancel_check_resource(self, ctxt, stack_id, engine_id):
        """Send check-resource cancel message.

        Sends a cancel message to given heat engine worker.
        """

        _client = messaging.get_rpc_client(
            topic=worker_api.TOPIC,
            version=self.BASE_RPC_API_VERSION,
            server=engine_id)

        method, kwargs = self.make_msg('cancel_check_resource',
                                       stack_id=stack_id)
        cl = _client.prepare(version='1.3')
        cl.cast(ctxt, method, **kwargs)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import template_format
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import stack as parser
from heat.engine import template
from heat.tests import common
from heat.tests import utils


rds_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "RDS Test",
  "Parameters" : {
    "KeyName" : {
      "Description" : "KeyName",
      "Type" : "String",
      "Default" : "test"
    }
   },
  "Resources" : {
    "DatabaseServer": {
      "Type": "AWS::RDS::DBInstance",
      "Properties": {
        "DBName"            : "wordpress",
        "Engine"            : "MySQL",
        "MasterUsername"    : "admin",
        "DBInstanceClass"   : "db.m1.small",
        "DBSecurityGroups"  : [],
        "AllocatedStorage"  : "5",
        "MasterUserPassword": "admin"
      }
    }
  }
}
'''


class DBInstance(resource.Resource):
    """Verify the schema of the new TemplateResource.

    This is copied from the old DBInstance.
    """
    properties_schema = {
        'DBSnapshotIdentifier': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'AllocatedStorage': properties.Schema(
            properties.Schema.STRING,
            required=True
        ),
        'AvailabilityZone': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'BackupRetentionPeriod': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'DBInstanceClass': properties.Schema(
            properties.Schema.STRING,
            required=True
        ),
        'DBName': properties.Schema(
            properties.Schema.STRING,
            required=False
        ),
        'DBParameterGroupName': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'DBSecurityGroups': properties.Schema(
            properties.Schema.LIST,
            required=False,
            default=[]
        ),
        'DBSubnetGroupName': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'Engine': properties.Schema(
            properties.Schema.STRING,
            constraints=[
                constraints.AllowedValues(['MySQL']),
            ],
            required=True
        ),
        'EngineVersion': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'LicenseModel': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'MasterUsername': properties.Schema(
            properties.Schema.STRING,
            required=True
        ),
        'MasterUserPassword': properties.Schema(
            properties.Schema.STRING,
            required=True
        ),
        'Port': properties.Schema(
            properties.Schema.STRING,
            required=False,
            default='3306'
        ),
        'PreferredBackupWindow': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'PreferredMaintenanceWindow': properties.Schema(
            properties.Schema.STRING,
            implemented=False
        ),
        'MultiAZ': properties.Schema(
            properties.Schema.BOOLEAN,
            implemented=False
        ),
    }

    # We only support a couple of the attributes right now
    attributes_schema = {
        "Endpoint.Address": attributes.Schema(
            "Connection endpoint for the database."
        ),
        "Endpoint.Port": attributes.Schema(
            ("The port number on which the database accepts "
             "connections.")
        ),
    }


class DBInstanceTest(common.HeatTestCase):
    def test_dbinstance(self):
        """Test that Template is parsable and publishes correct properties."""
        templ = template.Template(template_format.parse(rds_template))
        stack = parser.Stack(utils.dummy_context(), 'test_stack',
                             templ)

        res = stack['DatabaseServer']
        self.assertIsNone(res._validate_against_facade(DBInstance))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities to dynamically load plugin modules.

Modules imported this way remain accessible to static imports, regardless of
the order in which they are imported. For modules that are not part of an
existing package tree, use create_subpackage() to dynamically create a package
for them before loading them.
"""

import functools
import importlib.util
import pkgutil
import sys
import types

from oslo_log import log as logging


LOG = logging.getLogger(__name__)


def _module_name(*components):
    """Assemble a fully-qualified module name from its components."""
    return '.'.join(components)


def create_subpackage(path, parent_package_name, subpackage_name="plugins"):
    """Dynamically create a package into which to load plugins.

    This allows us to not include an __init__.py in the plugins directory. We
    must still create a package for plugins to go in, otherwise we get warning
    messages during import. This also provides a convenient place to store the
    path(s) to the plugins directory.
    """
    package_name = _module_name(parent_package_name, subpackage_name)

    package = types.ModuleType(package_name)
    package.__path__ = ([path] if isinstance(path, str)
                        else list(path))
    sys.modules[package_name] = package

    return package


def _import_module(importer, module_name, package):
    """Import a module dynamically into a package.

    :param importer: PEP302 Importer object (which knows the path to look in).
    :param module_name: the name of the module to import.
    :param package: the package to import the module into.
    """

    # Duplicate copies of modules are bad, so check if this has already been
    # imported statically
    if module_name in sys.modules:
        return sys.modules[module_name]

    module_spec = importer.find_spec(module_name)
    if module_spec is None:
        return None

    module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)

    # Make this accessible through the parent package for static imports
    local_name = module_name.partition(package.__name__ + '.')[2]
    module_components = local_name.split('.')
    parent = functools.reduce(getattr, module_components[:-1], package)
    setattr(parent, module_components[-1], module)

    return module


def load_modules(package, ignore_error=False):
    """Dynamically load all modules from a given package."""
    path = package.__path__
    pkg_prefix = package.__name__ + '.'

    for importer, module_name, is_package in pkgutil.walk_packages(path,
                                                                   pkg_prefix):
        # Skips tests or setup packages so as not to load tests in general
        # or setup.py during doc generation.
        if '.tests.' in module_name or module_name.endswith('.setup'):
            continue

        try:
            module = _import_module(importer, module_name, package)
        except ImportError:
            LOG.error('Failed to import module %s', module_name)
            if not ignore_error:
                raise
        else:
            if module is not None:
                yield module



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import weakref

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import reflection

from heat.common import exception
from heat.common.i18n import _
from heat.common import identifier
from heat.common import template_format
from heat.engine import attributes
from heat.engine import environment
from heat.engine import resource
from heat.engine import scheduler
from heat.engine import stack as parser
from heat.engine import stk_defn
from heat.engine import template
from heat.objects import raw_template
from heat.objects import stack as stack_object
from heat.objects import stack_lock
from heat.rpc import api as rpc_api

LOG = logging.getLogger(__name__)


class StackResource(resource.Resource):
    """Allows entire stack to be managed as a resource in a parent stack.

    An abstract Resource subclass that allows the management of an entire Stack
    as a resource in a parent stack.
    """

    # Assume True as this is evaluated before the stack is created
    # so there is no way to know for sure without subclass-specific
    # template parsing.
    requires_deferred_auth = True

    def __init__(self, name, json_snippet, stack):
        super(StackResource, self).__init__(name, json_snippet, stack)
        self._nested = None
        self._outputs = None
        self.resource_info = None

    def validate(self):
        super(StackResource, self).validate()
        # Don't redo a non-strict validation of a nested stack during the
        # creation of a child stack; only validate a child stack prior to the
        # creation of the root stack.
        if self.stack.nested_depth == 0 or not self.stack.strict_validate:
            self.validate_nested_stack()

    def validate_nested_stack(self):
        try:
            name = "%s-%s" % (self.stack.name, self.name)
            nested_stack = self._parse_nested_stack(
                name,
                self.child_template(),
                self.child_params())
            nested_stack.strict_validate = False
            nested_stack.validate()
        except AssertionError:
            raise
        except Exception as ex:
            path = "%s<%s>" % (self.name, self.template_url)
            raise exception.StackValidationFailed(
                error=ex, path=[self.stack.t.RESOURCES, path])

    @property
    def template_url(self):
        """Template url for the stack resource.

        When stack resource is a TemplateResource, it's the template
        location. For group resources like ResourceGroup where the
        template is constructed dynamically, it's just a placeholder.
        """

        return "nested_stack"

    def _outputs_to_attribs(self, json_snippet):
        outputs = json_snippet.get('Outputs')
        if not self.attributes and outputs:
            self.attributes_schema = (
                attributes.Attributes.schema_from_outputs(outputs))
            # Note: it can be updated too and for show return dictionary
            #       with all available outputs
            self.attributes = attributes.Attributes(
                self.name, self.attributes_schema,
                self._make_resolver(weakref.ref(self)))

    def _needs_update(self, after, before, after_props, before_props,
                      prev_resource, check_init_complete=True):

        # If the nested stack has not been created, use the default
        # implementation to determine if we need to replace the resource. Note
        # that we do *not* return the result.
        if self.resource_id is None:
            super(StackResource, self)._needs_update(after, before,
                                                     after_props, before_props,
                                                     prev_resource,
                                                     check_init_complete)
        else:
            if self.state == (self.CHECK, self.FAILED):
                nested_stack = self.rpc_client().show_stack(
                    self.context, self.nested_identifier())[0]
                nested_stack_state = (nested_stack[rpc_api.STACK_ACTION],
                                      nested_stack[rpc_api.STACK_STATUS])
                if nested_stack_state == (self.stack.CHECK, self.stack.FAILED):
                    # The stack-check action marked the stack resource
                    # CHECK_FAILED, so return True to allow the individual
                    # CHECK_FAILED resources decide if they need updating.
                    return True
                # The mark-unhealthy action marked the stack resource
                # CHECK_FAILED, so raise UpdateReplace to replace the
                # entire failed stack.
                raise resource.UpdateReplace(self)

        # Always issue an update to the nested stack and let the individual
        # resources in it decide if they need updating.
        return True

    def nested_identifier(self):
        if self.resource_id is None:
            return None
        return identifier.HeatIdentifier(
            self.context.project_id,
            self.physical_resource_name(),
            self.resource_id)

    def has_nested(self):
        """Return True if the resource has an existing nested stack."""
        return self.resource_id is not None or self._nested is not None

    def nested(self):
        """Return a Stack object representing the nested (child) stack.

        If we catch NotFound exception when loading, return None.
        """
        if self._nested is None and self.resource_id is not None:
            try:
                self._nested = parser.Stack.load(self.context,
                                                 self.resource_id)
            except exception.NotFound:
                return None

        return self._nested

    def child_template(self):
        """Default implementation to get the child template.

        Resources that inherit from StackResource should override this method
        with specific details about the template used by them.
        """
        raise NotImplementedError()

    def child_params(self):
        """Default implementation to get the child params.

        Resources that inherit from StackResource should override this method
        with specific details about the parameters used by them.
        """
        raise NotImplementedError()

    def preview(self):
        """Preview a StackResource as resources within a Stack.

        This method overrides the original Resource.preview to return a preview
        of all the resources contained in this Stack.  For this to be possible,
        the specific resources need to override both ``child_template`` and
        ``child_params`` with specific information to allow the stack to be
        parsed correctly. If any of these methods is missing, the entire
        StackResource will be returned as if it were a regular Resource.
        """
        try:
            child_template = self.child_template()
            params = self.child_params()
        except NotImplementedError:
            class_name = reflection.get_class_name(self, fully_qualified=False)
            LOG.warning("Preview of '%s' not yet implemented", class_name)
            return self

        name = "%s-%s" % (self.stack.name, self.name)
        self._nested = self._parse_nested_stack(name, child_template, params)

        return self.nested().preview_resources()

    def get_nested_parameters_stack(self):
        """Return a stack for schema validation.

        This returns a stack to be introspected for building parameters schema.
        It can be customized by subclass to return a restricted version of what
        will be running.
        """
        try:
            child_template = self.child_template()
            params = self.child_params()
        except NotImplementedError:
            class_name = reflection.get_class_name(self, fully_qualified=False)
            LOG.warning("Nested parameters of '%s' not yet "
                        "implemented", class_name)
            return
        name = "%s-%s" % (self.stack.name, self.name)
        return self._parse_nested_stack(name, child_template, params)

    def _parse_child_template(self, child_template, child_env):
        parsed_child_template = child_template
        if isinstance(parsed_child_template, template.Template):
            parsed_child_template = parsed_child_template.t
        return template.Template(parsed_child_template,
                                 files=self.child_template_files(child_env),
                                 env=child_env)

    def child_template_files(self, child_env):
        """Default implementation to get the files map for child template."""
        return self.stack.t.files

    def _parse_nested_stack(self, stack_name, child_template,
                            child_params, timeout_mins=None,
                            adopt_data=None):
        if timeout_mins is None:
            timeout_mins = self.stack.timeout_mins

        stack_user_project_id = self.stack.stack_user_project_id
        new_nested_depth = self._child_nested_depth()

        child_env = environment.get_child_environment(
            self.stack.env, child_params,
            child_resource_name=self.name,
            item_to_remove=self.resource_info)

        parsed_template = self._child_parsed_template(child_template,
                                                      child_env)
        self._validate_nested_resources(parsed_template)

        # Note we disable rollback for nested stacks, since they
        # should be rolled back by the parent stack on failure
        nested = parser.Stack(self.context,
                              stack_name,
                              parsed_template,
                              timeout_mins=timeout_mins,
                              disable_rollback=True,
                              parent_resource=self.name,
                              owner_id=self.stack.id,
                              user_creds_id=self.stack.user_creds_id,
                              stack_user_project_id=stack_user_project_id,
                              adopt_stack_data=adopt_data,
                              nested_depth=new_nested_depth)
        nested.set_parent_stack(self.stack)
        return nested

    def _child_nested_depth(self):
        if self.stack.nested_depth >= cfg.CONF.max_nested_stack_depth:
            msg = _("Recursion depth exceeds %d."
                    ) % cfg.CONF.max_nested_stack_depth
            raise exception.RequestLimitExceeded(message=msg)
        return self.stack.nested_depth + 1

    def _child_parsed_template(self, child_template, child_env):
        parsed_template = self._parse_child_template(child_template, child_env)

        # Don't overwrite the attributes_schema for subclasses that
        # define their own attributes_schema.
        if not hasattr(type(self), 'attributes_schema'):
            self.attributes = None
            self._outputs_to_attribs(parsed_template)
        return parsed_template

    def _validate_nested_resources(self, templ):
        if cfg.CONF.max_resources_per_stack == -1:
            return

        total_resources = (len(templ[templ.RESOURCES]) +
                           self.stack.total_resources(self.root_stack_id))

        identity = self.nested_identifier()
        if identity is not None:
            existing = self.rpc_client().list_stack_resources(self.context,
                                                              identity)
            # Don't double-count existing resources during an update
            total_resources -= len(existing)

        if (total_resources > cfg.CONF.max_resources_per_stack):
            message = exception.StackResourceLimitExceeded.msg_fmt
            raise exception.RequestLimitExceeded(message=message)

    def create_with_template(self, child_template, user_params=None,
                             timeout_mins=None, adopt_data=None):
        """Create the nested stack with the given template."""
        name = self.physical_resource_name()
        if timeout_mins is None:
            timeout_mins = self.stack.timeout_mins
        stack_user_project_id = self.stack.stack_user_project_id

        kwargs = self._stack_kwargs(user_params, child_template, adopt_data)

        adopt_data_str = None
        if adopt_data is not None:
            if 'environment' not in adopt_data:
                adopt_data['environment'] = kwargs['params']
            if 'template' not in adopt_data:
                if isinstance(child_template, template.Template):
                    adopt_data['template'] = child_template.t
                else:
                    adopt_data['template'] = child_template
            adopt_data_str = json.dumps(adopt_data)

        args = {rpc_api.PARAM_TIMEOUT: timeout_mins,
                rpc_api.PARAM_DISABLE_ROLLBACK: True,
                rpc_api.PARAM_ADOPT_STACK_DATA: adopt_data_str}
        kwargs.update({
            'stack_name': name,
            'args': args,
            'environment_files': None,
            'owner_id': self.stack.id,
            'user_creds_id': self.stack.user_creds_id,
            'stack_user_project_id': stack_user_project_id,
            'nested_depth': self._child_nested_depth(),
            'parent_resource_name': self.name
        })
        with self.translate_remote_exceptions:
            try:
                result = self.rpc_client()._create_stack(self.context,
                                                         **kwargs)
            except exception.HeatException:
                with excutils.save_and_reraise_exception():
                    if adopt_data is None:
                        raw_template.RawTemplate.delete(self.context,
                                                        kwargs['template_id'])

        self.resource_id_set(result['stack_id'])

    def child_definition(self, child_template=None, user_params=None,
                         nested_identifier=None):
        if user_params is None:
            user_params = self.child_params()
        if child_template is None:
            child_template = self.child_template()
        if nested_identifier is None:
            nested_identifier = self.nested_identifier()

        child_env = environment.get_child_environment(
            self.stack.env,
            user_params,
            child_resource_name=self.name,
            item_to_remove=self.resource_info)

        parsed_template = self._child_parsed_template(child_template,
                                                      child_env)
        return stk_defn.StackDefinition(self.context, parsed_template,
                                        nested_identifier,
                                        None)

    def _stack_kwargs(self, user_params, child_template, adopt_data=None):
        defn = self.child_definition(child_template, user_params)
        parsed_template = defn.t

        if adopt_data is None:
            template_id = parsed_template.store(self.context)
            return {
                'template_id': template_id,
                'template': None,
                'params': None,
                'files': None,
            }
        else:
            return {
                'template': parsed_template.t,
                'params': defn.env.user_env_as_dict(),
                'files': parsed_template.files,
            }

    @excutils.exception_filter
    def translate_remote_exceptions(self, ex):
        if (isinstance(ex, exception.ActionInProgress) and
                self.stack.action == self.stack.ROLLBACK):
            # The update was interrupted and the rollback is already in
            # progress, so just ignore the error and wait for the rollback to
            # finish
            return True

        class_name = reflection.get_class_name(ex, fully_qualified=False)
        if not class_name.endswith('_Remote'):
            return False

        full_message = str(ex)
        if full_message.find('\n') > -1:
            message, msg_trace = full_message.split('\n', 1)
        else:
            message = full_message

        raise exception.ResourceFailure(message, self, action=self.action)

    def check_create_complete(self, cookie=None):
        return self._check_status_complete(self.CREATE)

    def _check_status_complete(self, expected_action, cookie=None):

        try:
            data = stack_object.Stack.get_status(self.context,
                                                 self.resource_id)
        except exception.NotFound:
            if expected_action == self.DELETE:
                return True
            # It's possible the engine handling the create hasn't persisted
            # the stack to the DB when we first start polling for state
            return False

        action, status, status_reason, updated_time = data

        if action != expected_action:
            return False

        if status == self.IN_PROGRESS:
            if cookie is not None and 'fail_count' in cookie:
                prev_status_reason = cookie['previous']['status_reason']
                if status_reason != prev_status_reason:
                    # State has changed, so fail on the next failure
                    cookie['fail_count'] = 1
            return False
        elif status == self.COMPLETE:
            # For operations where we do not take a resource lock
            # (i.e. legacy-style), check that the stack lock has been
            # released before reporting completeness.
            done = (self._should_lock_on_action(expected_action) or
                    stack_lock.StackLock.get_engine_id(
                self.context, self.resource_id) is None)
            if done:
                # Reset nested, to indicate we changed status
                self._nested = None
            return done
        elif status == self.FAILED:
            if cookie is not None and 'fail_count' in cookie:
                cookie['fail_count'] -= 1
                if cookie['fail_count'] > 0:
                    raise resource.PollDelay(10)
            raise exception.ResourceFailure(status_reason, self,
                                            action=action)
        else:
            raise exception.ResourceUnknownStatus(
                resource_status=status,
                status_reason=status_reason,
                result=_('Stack unknown status'))

    def check_adopt_complete(self, cookie=None):
        return self._check_status_complete(self.ADOPT)

    def _try_rollback(self):
        stack_identity = self.nested_identifier()
        if stack_identity is None:
            return False

        try:
            self.rpc_client().stack_cancel_update(
                self.context,
                dict(stack_identity),
                cancel_with_rollback=True)
        except exception.NotSupported:
            return False

        try:
            data = stack_object.Stack.get_status(self.context,
                                                 self.resource_id)
        except exception.NotFound:
            return False

        action, status, status_reason, updated_time = data

        # If nested stack is still in progress, it should eventually roll
        # itself back due to stack_cancel_update(), so we just need to wait
        # for that to complete
        return status == self.stack.IN_PROGRESS

    def update_with_template(self, child_template, user_params=None,
                             timeout_mins=None):
        """Update the nested stack with the new template."""
        if self.id is None:
            self.store()

        if self.stack.action == self.stack.ROLLBACK:
            if self._try_rollback():
                LOG.info('Triggered nested stack %s rollback',
                         self.physical_resource_name())
                return {'target_action': self.stack.ROLLBACK}

        if self.resource_id is None:
            # if the create failed for some reason and the nested
            # stack was not created, we need to create an empty stack
            # here so that the update will work.
            def _check_for_completion():
                while not self.check_create_complete():
                    yield

            empty_temp = template_format.parse(
                "heat_template_version: '2013-05-23'")
            self.create_with_template(empty_temp, {})
            checker = scheduler.TaskRunner(_check_for_completion)
            checker(timeout=self.stack.timeout_secs())

        if timeout_mins is None:
            timeout_mins = self.stack.timeout_mins

        try:
            status_data = stack_object.Stack.get_status(self.context,
                                                        self.resource_id)
        except exception.NotFound:
            raise resource.UpdateReplace(self)

        action, status, status_reason, updated_time = status_data

        kwargs = self._stack_kwargs(user_params, child_template)

        kwargs.update({
            'stack_identity': dict(self.nested_identifier()),
            'args': {rpc_api.PARAM_TIMEOUT: timeout_mins,
                     rpc_api.PARAM_CONVERGE: self.converge}
        })
        with self.translate_remote_exceptions:
            try:
                self.rpc_client()._update_stack(self.context, **kwargs)
            except exception.HeatException:
                with excutils.save_and_reraise_exception():
                    raw_template.RawTemplate.delete(self.context,
                                                    kwargs['template_id'])

    def check_update_complete(self, cookie=None):
        if cookie is not None and 'target_action' in cookie:
            target_action = cookie['target_action']
            cookie = None
        else:
            target_action = self.stack.UPDATE
        return self._check_status_complete(target_action,
                                           cookie=cookie)

    def _handle_cancel(self):
        stack_identity = self.nested_identifier()
        if stack_identity is not None:
            LOG.debug('Cancelling %s of %s' % (self.action, self))
            try:
                self.rpc_client().stack_cancel_update(
                    self.context,
                    dict(stack_identity),
                    cancel_with_rollback=False)
            except exception.NotSupported:
                LOG.debug('Nested stack %s not in cancellable state',
                          stack_identity.stack_name)

    def handle_preempt(self):
        self._handle_cancel()

    def handle_update_cancel(self, cookie):
        self._handle_cancel()

    def handle_create_cancel(self, cookie):
        return self.handle_update_cancel(cookie)

    def delete_nested(self):
        """Delete the nested stack."""
        stack_identity = self.nested_identifier()
        if stack_identity is None:
            return

        cookie = None
        if not self.stack.convergence:
            try:
                status_data = stack_object.Stack.get_status(self.context,
                                                            self.resource_id)
            except exception.NotFound:
                return

            action, status, status_reason, updated_time = status_data
            if (action, status) == (self.stack.DELETE,
                                    self.stack.IN_PROGRESS):
                cookie = {
                    'previous': {
                        'state': (action, status),
                        'status_reason': status_reason,
                        'updated_at': None,
                    },
                    'fail_count': 2,
                }

        with self.rpc_client().ignore_error_by_name('EntityNotFound'):
            if self.abandon_in_progress:
                self.rpc_client().abandon_stack(self.context, stack_identity)
            else:
                self.rpc_client().delete_stack(self.context, stack_identity,
                                               cast=False)
            return cookie

    def handle_delete(self):
        return self.delete_nested()

    def check_delete_complete(self, cookie=None):
        return self._check_status_complete(self.DELETE)

    def handle_suspend(self):
        stack_identity = self.nested_identifier()
        if stack_identity is None:
            raise exception.Error(_('Cannot suspend %s, stack not created')
                                  % self.name)
        self.rpc_client().stack_suspend(self.context, dict(stack_identity))

    def check_suspend_complete(self, cookie=None):
        return self._check_status_complete(self.SUSPEND)

    def handle_resume(self):
        stack_identity = self.nested_identifier()
        if stack_identity is None:
            raise exception.Error(_('Cannot resume %s, stack not created')
                                  % self.name)
        self.rpc_client().stack_resume(self.context, dict(stack_identity))

    def check_resume_complete(self, cookie=None):
        return self._check_status_complete(self.RESUME)

    def handle_check(self):
        stack_identity = self.nested_identifier()
        if stack_identity is None:
            raise exception.Error(_('Cannot check %s, stack not created')
                                  % self.name)

        self.rpc_client().stack_check(self.context, dict(stack_identity))

    def check_check_complete(self, cookie=None):
        return self._check_status_complete(self.CHECK)

    def prepare_abandon(self):
        self.abandon_in_progress = True
        nested_stack = self.nested()
        if nested_stack:
            return nested_stack.prepare_abandon()

        return {}

    def get_output(self, op):
        """Return the specified Output value from the nested stack.

        If the output key does not exist, raise a NotFound exception.
        """
        if (self._outputs is None or
                (op in self._outputs and
                 rpc_api.OUTPUT_ERROR not in self._outputs[op] and
                 self._outputs[op].get(rpc_api.OUTPUT_VALUE) is None)):
            stack_identity = self.nested_identifier()
            if stack_identity is None:
                return
            stack = self.rpc_client().show_stack(self.context,
                                                 dict(stack_identity))
            if not stack:
                return
            outputs = stack[0].get(rpc_api.STACK_OUTPUTS) or {}
            self._outputs = {o[rpc_api.OUTPUT_KEY]: o for o in outputs}

        if op not in self._outputs:
            raise exception.NotFound(_('Specified output key %s not '
                                       'found.') % op)

        output_data = self._outputs[op]
        if rpc_api.OUTPUT_ERROR in output_data:
            raise exception.TemplateOutputError(
                resource=self.name,
                attribute=op,
                message=output_data[rpc_api.OUTPUT_ERROR])

        return output_data[rpc_api.OUTPUT_VALUE]

    def _resolve_attribute(self, name):
        try:
            return self.get_output(name)
        except exception.NotFound:
            raise exception.InvalidTemplateAttribute(resource=self.name,
                                                     key=name)
